from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np

import os
import time
import multiprocessing as mp
from . import mcfm_parallel_worker as mpw


def get_cluster_data(atoms=None,
                     clusterData=None,
                     mcfm_pot=None):
    """Obtain a list of cluster data with calculations being done in parallel

    Parameters
    ----------
    atoms : ase.Atoms
        atoms object representing the structure
    clusterData : list
        List of empty objects to be filled with clusterData instances
    mcfm_pot : matscipy.calculators.mcfm.MultiClusterForceMixing
        qmmm potential
    """
    # number of porcessors
    try:
        nProc = int(os.environ["OMP_NUM_THREADS"])
    except KeyError:
        nProc = mp.cpu_count() / 2

    # number of threads - number of clusters
    numThreads = len(mcfm_pot.cluster_list)

    # In case there are not enough cpu's,
    # have the number of processes artificially increased
    if (numThreads > nProc):
        nProc = numThreads

    # Create atomic clusters and evaluate their sizes
    atomicClustersList = []

    for cluster in mcfm_pot.cluster_list:
        atomicCluster = mcfm_pot.qm_cluster.carve_cluster(atoms,
                                                          cluster,
                                                          buffer_hops=mcfm_pot.buffer_hops)
        atomicClustersList.append(atomicCluster)

    # ------ Evaluate work balancing
    valenceElectrons = [np.sum(np.abs(item.numbers - 2)) for item in atomicClustersList]
    fractionWorkloadPerCluster = [(item ** 2) for item in valenceElectrons]
    totalWorkload = sum(fractionWorkloadPerCluster)
    fractionWorkloadPerCluster = [item / totalWorkload for item in fractionWorkloadPerCluster]

    nProcPerCluster = [1 for item in atomicClustersList]
    leftoverProcs = nProc - sum(nProcPerCluster)
    # Distribute leftoverProcs
    for i in range(numThreads):
        nProcPerCluster[i] += int(fractionWorkloadPerCluster[i] * leftoverProcs)

    # Disribute leftover procs (if any)
    leftoverProcs = nProc - sum(nProcPerCluster)
    running = True
    while running:
        for i in np.argsort(fractionWorkloadPerCluster)[::-1]:
            if (leftoverProcs <= 0):
                running = False
                break
            nProcPerCluster[i] += 1
            leftoverProcs -= 1

    if (mcfm_pot.debug_qm_calculator):
        print(fractionWorkloadPerCluster, nProcPerCluster, ":parallelTime")

    # Set up the Manager
    mpManager = mp.Manager()
    sharedList = mpManager.list(list(range(numThreads)))

    # Setup a list of processes that we want to run
    processes = []
    for rank in range(numThreads):
        p = mp.Process(target=mpw.worker_populate_cluster_data,
                       name=None,
                       args=(rank, numThreads),
                       kwargs=dict(nProcLocal=nProcPerCluster[rank],
                                   atomic_cluster=atomicClustersList[rank],
                                   clusterIndexes=mcfm_pot.cluster_list[rank],
                                   nAtoms=len(atoms),
                                   qmCalculator=mcfm_pot.qm_calculator,
                                   sharedList=sharedList,
                                   debug_qm_calculator=mcfm_pot.debug_qm_calculator))
        processes.append(p)

    # Run processes
    for p in processes:
        p.start()
        # Each QM calculation takes between 1 and 100s so the wait shouldnt affect the performance,
        # It helps prevent any I/O clashes when setting up the simulations
        time.sleep(1e-3)

    # Exit the completed processes
    for p in processes:
        p.join()

    # Extract results
    for index in range(len(clusterData)):
        clusterData[index] = sharedList[index]



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

import itertools

import numpy as np

import ase

from matscipy.neighbours import first_neighbours, neighbour_list

###

def hydrogenate(a, cutoff, bond_length, b=None, mask=[True, True, True],
                exclude=None, vacuum=None):
    """
    Hydrogenate a slab of material at its periodic boundary conditions.
    Boundary conditions are turned into nonperiodic.

    Parameters
    ----------
    a : ase.Atoms
        Atomic configuration.
    cutoff : float
        Cutoff for neighbor counting.
    bond_length : float
        X-H bond length for hydrogenation.
    b : ase.Atoms, optional
        If present, this is the configuration to hydrogenate. Number of atoms
        must be identical to a object. All bonds present in a but not present
        in b will be hydrogenated in b.
    mask : list of bool
        Cartesian directions which to hydrogenate, only if b argument is not
        given.
    exclude : array_like
        Boolean array masking atoms to be excluded from hydrogenation.
    vacuum : float, optional
        Add this much vacuum after hydrogenation.

    Returns
    -------
    a : ase.Atoms
        Atomic configuration of the hydrogenated slab.
    """
    if b is None:
        b = a.copy()
        b.set_pbc(np.logical_not(mask))

    if exclude is None:
        exclude = np.zeros(len(a), dtype=bool)

    i_a, j_a, D_a, d_a = neighbour_list('ijDd', a, cutoff)
    i_b, j_b = neighbour_list('ij', b, cutoff)

    firstneigh_a = first_neighbours(len(a), i_a)
    firstneigh_b = first_neighbours(len(b), i_b)

    coord_a = np.bincount(i_a, minlength=len(a))
    coord_b = np.bincount(i_b, minlength=len(b))

    hydrogens = []
    # Surface atoms have coord_a != coord_b. Those need hydrogenation
    for k in np.arange(len(a))[np.logical_and(coord_a!=coord_b,
                                              np.logical_not(exclude))]:
        l1_a = firstneigh_a[k]
        l2_a = firstneigh_a[k+1]
        l1_b = firstneigh_b[k]
        l2_b = firstneigh_b[k+1]
        n_H = 0
        for l_a in range(l1_a, l2_a):
            assert i_a[l_a] == k
            bond_exists = False
            for l_b in range(l1_b, l2_b):
                assert i_b[l_b] == k
                if j_a[l_a] == j_b[l_b]:
                    bond_exists = True
            if not bond_exists:
                # Bond existed before cut
                hydrogens += [b[k].position+bond_length*D_a[l_a]/d_a[l_a]]
                n_H += 1
        assert n_H == coord_a[k]-coord_b[k]

    if hydrogens == []:
        raise RuntimeError('No Hydrogen created.')

    b += ase.Atoms(['H']*len(hydrogens), hydrogens)

    if vacuum is not None:
        axis=[]
        for i in range(3):
            if mask[i]:
                axis += [i]
        b.center(vacuum, axis=axis)

    return b



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2016) Andreas Klemenz, Fraunhofer IWM
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

import time
import sys
import distutils.version

import numpy as np
import ase
import ase.data
import ase.io
import ase.io.lammpsrun
import ase.calculators.lammpsrun
import matscipy.neighbours

import matscipy.opls

try:
    import ase.version
    ase_version_str = ase.version.version
except:
    ase_version_str = ase.__version__


def read_extended_xyz(fileobj):
    """Read extended xyz file with labeled atoms."""
    atoms = ase.io.read(fileobj)
    opls_struct = matscipy.opls.OPLSStructure(atoms)
    opls_struct.arrays = atoms.arrays

    types = opls_struct.get_array('type')
    opls_struct.types = np.unique(types)

    tags = np.zeros(len(opls_struct), dtype=int)
    for it, type in enumerate(opls_struct.types):
        tags[types == type] = it
    opls_struct.set_tags(tags)

    return opls_struct


def read_block(filename, name):
    data = {}
    if isinstance(filename, str):
        fileobj = open(filename, 'r')

    block = False
    for line in fileobj.readlines():
        line = line.split()

        # find data block
        if len(line) >= 2:
            if line[1] == name:
                block = True

        # end of data block
        if block == True and len(line) == 0:
            block = False

        # read data
        if block:
            if line[0][0] == '#':
                continue
            else:
                symbol = line[0]
                data[symbol] = []
                for word in line[1:]:
                    if word[0] == '#':
                        break
                    else:
                        data[symbol].append(float(word))
                if len(data[symbol]) == 1:
                    data[symbol] = data[symbol][0]

    if len(data) == 0:
        print('Error: Data block \"%s\" not found in file \"%s\"' % (name, filename))
        sys.exit()

    fileobj.close()
    return data


def read_cutoffs(filename):
    cutoffs = matscipy.opls.CutoffList(read_block(filename, 'Cutoffs'))
    return cutoffs


def read_parameter_file(filename):
    one       = read_block(filename, 'Element')
    bonds     = matscipy.opls.BondData(read_block(filename, 'Bonds'))
    angles    = matscipy.opls.AnglesData(read_block(filename, 'Angles'))
    dihedrals = matscipy.opls.DihedralsData(read_block(filename, 'Dihedrals'))
    cutoffs   = matscipy.opls.CutoffList(read_block(filename, 'Cutoffs'))

    return cutoffs, one, bonds, angles, dihedrals

def write_lammps(prefix, atoms):
    write_lammps_in(prefix)
    write_lammps_atoms(prefix, atoms)
    write_lammps_definitions(prefix, atoms)

def write_lammps_in(prefix):
    if isinstance(prefix, str):
        fileobj = open(prefix + '.in', 'w')
    fileobj.write("""# LAMMPS relaxation (written by ASE)

units           metal
atom_style      full
boundary        p p p
#boundary       p p f

""")
    fileobj.write('read_data ' + prefix + '.atoms\n')
    fileobj.write('include  ' + prefix + '.opls\n')
    fileobj.write("""
kspace_style    pppm 1e-5
#kspace_modify  slab 3.0

neighbor        1.0 bin
neigh_modify    delay 0 every 1 check yes

thermo          1000
thermo_style    custom step temp press cpu pxx pyy pzz pxy pxz pyz ke pe etotal vol lx ly lz atoms

dump            1 all xyz 1000 dump_relax.xyz
dump_modify     1 sort id

restart         100000 test_relax

min_style       fire
minimize        1.0e-14 1.0e-5 100000 100000
""")
    fileobj.close()


def write_lammps_atoms(prefix, atoms):
    """Write atoms input for LAMMPS"""
    if isinstance(prefix, str):
        fileobj = open(prefix + '.atoms', 'w')

    # header
    fileobj.write(fileobj.name + ' (by write_lammps_atoms)\n\n')
    fileobj.write(str(len(atoms)) + ' atoms\n')
    fileobj.write(str(len(atoms.types)) + ' atom types\n')

    blist = atoms.bond_list
    if len(blist):
        btypes = atoms.bond_types
        fileobj.write(str(len(blist)) + ' bonds\n')
        fileobj.write(str(len(btypes)) + ' bond types\n')

    alist = atoms.ang_list
    if len(alist):
        atypes = atoms.ang_types
        fileobj.write(str(len(alist)) + ' angles\n')
        fileobj.write(str(len(atypes)) + ' angle types\n')

    dlist = atoms.dih_list
    if len(dlist):
        dtypes = atoms.dih_types
        fileobj.write(str(len(dlist)) + ' dihedrals\n')
        fileobj.write(str(len(dtypes)) + ' dihedral types\n')

    # cell
    if distutils.version.LooseVersion(ase_version_str) > distutils.version.LooseVersion('3.11.0'):
        p = ase.calculators.lammpsrun.Prism(atoms.get_cell())
    else:
        p = ase.calculators.lammpsrun.prism(atoms.get_cell())

    xhi, yhi, zhi, xy, xz, yz = p.get_lammps_prism_str()
    fileobj.write('\n0.0 %s  xlo xhi\n' % xhi)
    fileobj.write('0.0 %s  ylo yhi\n' % yhi)
    fileobj.write('0.0 %s  zlo zhi\n' % zhi)

    
    # atoms
    fileobj.write('\nAtoms\n\n')
    tags  = atoms.get_tags()
    types = atoms.types
    if atoms.has('molid'):
        molid = atoms.get_array('molid')
    else:
        molid = [1] * len(atoms)

    if distutils.version.LooseVersion(ase_version_str) > distutils.version.LooseVersion('3.13.0'):
        positions_lammps_str = p.positions_to_lammps_strs(atoms.get_positions())
    else:
        positions_lammps_str = map(p.pos_to_lammps_str, atoms.get_positions())

    for i, r in enumerate(positions_lammps_str):
        q = atoms.atom_data[types[tags[i]]][2]
        fileobj.write('%6d %3d %3d %s %s %s %s' % ((i + 1, molid[i],
                                                    tags[i] + 1, 
                                                    q)
                                                   + tuple(r)))
        fileobj.write(' # ' + atoms.types[tags[i]] + '\n')

    # velocities
    velocities = atoms.get_velocities()
    if velocities is not None:
        fileobj.write('\nVelocities\n\n')
        for i, v in enumerate(velocities):
            fileobj.write('%6d %g %g %g\n' %
                          (i + 1, v[0], v[1], v[2]))

    # masses
    masses = atoms.get_masses()
    tags   = atoms.get_tags()

    fileobj.write('\nMasses\n\n')
    for i, type, tag in zip(range(len(atoms.types)), atoms.types, np.unique(tags)):
        fileobj.write('%6d %g # %s\n' % 
                      (i + 1, 
                       masses[tags == tag][0],
                       type))

    # bonds
    if len(blist):
        fileobj.write('\nBonds\n\n')
        for ib, bvals in enumerate(blist):
            fileobj.write('%8d %6d %6d %6d ' %
                          (ib + 1, bvals[0] + 1, bvals[1] + 1, 
                           bvals[2] + 1))
            try:
                fileobj.write('# ' + btypes[bvals[0]])
            except:
                pass
            fileobj.write('\n')

    # angles
    if len(alist):
        fileobj.write('\nAngles\n\n')
        for ia, avals in enumerate(alist):
            fileobj.write('%8d %6d %6d %6d %6d ' %
                          (ia + 1, avals[0] + 1, 
                           avals[1] + 1, avals[2] + 1, avals[3] + 1))
            try:
                fileobj.write('# ' + atypes[avals[0]])
            except:
                pass
            fileobj.write('\n')

    # dihedrals
    if len(dlist):
        fileobj.write('\nDihedrals\n\n')
        for i, dvals in enumerate(dlist):
            fileobj.write('%8d %6d %6d %6d %6d %6d ' %
                          (i + 1, dvals[0] + 1, 
                           dvals[1] + 1, dvals[2] + 1, 
                           dvals[3] + 1, dvals[4] + 1))
            try:
                fileobj.write('# ' + dtypes[dvals[0]])
            except:
                pass
            fileobj.write('\n')


def write_lammps_definitions(prefix, atoms):
    """Write force field definitions for LAMMPS."""

    if isinstance(prefix, str):
        fileobj = open(prefix + '.opls', 'w')

    fileobj.write('# OPLS potential\n')
    fileobj.write('# write_lammps ' +
                  str(time.asctime(
                time.localtime(time.time()))))

    # bonds
    if len(atoms.bond_types):
        fileobj.write('\n# bonds\n')
        fileobj.write('bond_style      harmonic\n')
        for ib, btype in enumerate(atoms.bond_types):
            fileobj.write('bond_coeff %6d' % (ib + 1))
            itype, jtype = btype.split('-')
            name, values = atoms.bonds.name_value(itype, jtype)
            for value in values:
                fileobj.write(' ' + str(value))
            fileobj.write(' # ' + name + '\n')

    # angles
    if len(atoms.ang_types):
        fileobj.write('\n# angles\n')
        fileobj.write('angle_style      harmonic\n')
        for ia, atype in enumerate(atoms.ang_types):
            fileobj.write('angle_coeff %6d' % (ia + 1))
            itype, jtype, ktype = atype.split('-')
            name, values = atoms.angles.name_value(itype, jtype, ktype)
            for value in values:
                fileobj.write(' ' + str(value))
            fileobj.write(' # ' + name + '\n')


    # dihedrals
    if len(atoms.dih_types):
        fileobj.write('\n# dihedrals\n')
        fileobj.write('dihedral_style      opls\n')
        for id, dtype in enumerate(atoms.dih_types):
            fileobj.write('dihedral_coeff %6d' % (id + 1))
            itype, jtype, ktype, ltype = dtype.split('-')
            name, values = atoms.dihedrals.name_value(itype, jtype, ktype, ltype)
            for value in values:
                fileobj.write(' ' + str(value))
            fileobj.write(' # ' + name + '\n')


    # Lennard Jones settings
    fileobj.write('\n# L-J parameters\n')
    fileobj.write('pair_style lj/cut/coul/long 10.0 7.4' +
                  ' # consider changing these parameters\n')
    fileobj.write('special_bonds lj/coul 0.0 0.0 0.5\n')
    for ia, atype in enumerate(atoms.types):
        if len(atype) < 2:
            atype = atype + ' '
        fileobj.write('pair_coeff ' + str(ia + 1) + ' ' + str(ia + 1))
        for value in atoms.atom_data[atype][:2]:
            fileobj.write(' ' + str(value))
        fileobj.write(' # ' + atype + '\n')
    fileobj.write('pair_modify shift yes mix geometric\n')


    # Charges
    fileobj.write('\n# charges\n')
    for ia, atype in enumerate(atoms.types):
        if len(atype) < 2:
            atype = atype + ' '
        fileobj.write('set type ' + str(ia + 1))
        fileobj.write(' charge ' + str(atoms.atom_data[atype][2]))
        fileobj.write(' # ' + atype + '\n')


def read_lammps_data(filename):
    """Read positions, connectivities, etc."""
    
    if isinstance(filename, str):
        fileobj = open(filename, 'r')

    lines = fileobj.readlines()
    lines.pop(0)

    def next_entry():
        line = lines.pop(0).strip()
        if(len(line) > 0):
            lines.insert(0, line)

    def next_key():
        while(len(lines)):
            line = lines.pop(0).strip()
            if(len(line) > 0):
                lines.pop(0)
                return line
        return None

    next_entry()
    header = {}
    while(True):
        line = lines.pop(0).strip()
        if len(line):
            w = line.split()
            if len(w) == 2:
                header[w[1]] = int(w[0])
            else:
                header[w[1] + ' ' + w[2]] = int(w[0])
        else:
            break


    # read box
    next_entry()
    cell = np.zeros(3)
    for i in range(3):
        line = lines.pop(0).strip()
        cell[i] = float(line.split()[1])


    while(not lines.pop(0).startswith('Atoms')):
        pass
    lines.pop(0)


    natoms = header['atoms']
    molid     = np.ones(natoms, dtype=int)
    tags      = np.ones(natoms, dtype=int)
    charges   = np.zeros(natoms, dtype=float)
    positions = np.zeros([natoms,3])
    types     = ['']*header['atom types']
    inconsistent = False

    for line in lines[:natoms]:
        w = line.split()
        i = int(w[0])-1
        molid[i]        = int(w[1])
        tags[i]         = int(w[2])-1
        charges[i]      = float(w[3])
        positions[i][0] = float(w[4])
        positions[i][1] = float(w[5])
        positions[i][2] = float(w[6])

        # try to read atom type from comment
        if len(w) >= 8:
            type = ''.join(w[8:])
            if types[tags[i]] == type:
                pass
            elif types[tags[i]] == '':
                types[tags[i]] = type
            else:
                inconsistent = True

    if inconsistent:
        print('WARNING: Inconsistency between particle descriptions and particle tags found.')
        types = []
        for type in np.unique(tags):
            types.append(str(type))


    opls_struct = matscipy.opls.OPLSStructure(str(natoms)+'H', positions=positions, cell=cell)
    opls_struct.set_tags(tags)
    opls_struct.set_array('molid', molid)

    opls_struct.atom_data = {}
    opls_struct.types = types
    for tag, type in zip(np.unique(tags), types):
        opls_struct.atom_data[type] = [0.0, 0.0, charges[tags == tag][0]]

    del lines[:natoms]

    key = next_key()


    velocities = np.zeros([natoms,3])
    if key == 'Velocities':
        for line in lines[:natoms]:
            w = line.split()
            i = int(w[0])-1
            velocities[i][0] = float(w[1])
            velocities[i][1] = float(w[2])
            velocities[i][2] = float(w[3])

        del lines[:natoms]

        key = next_key()


    if key == 'Masses':
        ntypes = len(opls_struct.atom_data)
        masses = np.empty((ntypes))
        for line in lines[:ntypes]:
            w = line.split()
            i = int(w[0])-1
            masses[i] = float(w[1])

        del lines[:ntypes]

        opls_struct.set_masses(masses[tags])
        opls_struct.set_velocities(velocities)

        # get the elements from the masses
        # this ensures that we have the right elements
        # even when reading from a lammps dump file
        def newtype(element, types):
            if len(element) > 1:
                # can not extend, we are restricted to
                # two characters
                return element
            count = 0
            for type in types:
                if type[0] == element:
                    count += 1
            label = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'
            return (element + label[count])

        atomic_numbers = np.empty(ntypes, dtype=int)
        ams = ase.data.atomic_masses[:]
        ams[np.isnan(ams)] = 0
        for i, mass in enumerate(masses):
            m2 = (ams - mass)**2
            atomic_numbers[i] = m2.argmin()

        opls_struct.set_atomic_numbers(atomic_numbers[tags])

        key = next_key()
                


    if key != 'Bonds':
        bond_list = np.empty([0,3], dtype=int)
        bond_types = np.empty(0, dtype=str)
    else:
        nbonds = header['bonds']
        bond_list  = np.empty([nbonds,3], dtype=int)
        bond_types = ['']*header['bond types']
        inconsistent = False

        for line in lines[:nbonds]:
            w = line.split()
            i = int(w[0])-1
            bond_list[i][0] = int(w[1])-1
            bond_list[i][1] = int(w[2])-1
            bond_list[i][2] = int(w[3])-1

            # try to read bond type info from comment
            if len(w) >= 5:
                bond_type = ''.join(w[5:])
                if bond_types[bond_list[i][0]-1] == bond_type:
                    pass
                elif bond_types[bond_list[i][0]-1] == '':
                    bond_types[bond_list[i][0]-1] = bond_type
                else:
                    inconsistent = True

        if inconsistent:
            print('WARNING: Inconsistency between bond descriptions and bond type numbers found.')
            bond_types = ['']*header['bond types']

        del lines[:nbonds]

        key = next_key()

    opls_struct.bond_types = bond_types
    opls_struct.bond_list = bond_list



    if key != 'Angles':
        ang_list = np.empty([0,4], dtype=int)
        ang_types = np.empty(0, dtype=str)
    else:
        nangles = header['angles']
        ang_list  = np.empty([nangles,4], dtype=int)
        ang_types = ['']*header['angle types']
        inconsistent = False

        for line in lines[:nangles]:
            w = line.split()
            i = int(w[0])-1
            ang_list[i][0] = int(w[1])-1
            ang_list[i][1] = int(w[2])-1
            ang_list[i][2] = int(w[3])-1
            ang_list[i][3] = int(w[4])-1

            # try to read angle type info from comment
            if len(w) >= 5:
                ang_type = ''.join(w[6:])
                if ang_types[ang_list[i][0]-1] == ang_type:
                    pass
                elif ang_types[ang_list[i][0]-1] == '':
                    ang_types[ang_list[i][0]-1] = ang_type
                else:
                    inconsistent = True

        if inconsistent:
            print('WARNING: Inconsistency between angle descriptions and angle type numbers found.')
            ang_types = ['']*header['angle types']

        del lines[:nangles]

        key = next_key()

    opls_struct.ang_types = ang_types
    opls_struct.ang_list = ang_list



    if key != 'Dihedrals':
        dih_list = np.empty([0,5], dtype=int)
        dih_types = np.empty(header['dihedral types'], dtype=str)
    else:
        ndihedrals = header['dihedrals']
        dih_list = np.empty([ndihedrals,5], dtype=int)
        dih_types = ['']*header['dihedral types']
        inconsistent = False

        for line in lines[:ndihedrals]:
            w = line.split()
            i = int(w[0])-1
            dih_list[i][0] = int(w[1])-1
            dih_list[i][1] = int(w[2])-1
            dih_list[i][2] = int(w[3])-1
            dih_list[i][3] = int(w[4])-1
            dih_list[i][4] = int(w[5])-1

            # try to read dihedral type info from comment
            if len(w) >= 7:
                dih_type = ''.join(w[7:])
                if dih_types[dih_list[i][0]-1] == dih_type:
                    pass
                elif dih_types[dih_list[i][0]-1] == '':
                    dih_types[dih_list[i][0]-1] = dih_type
                else:
                    inconsistent = True

        if inconsistent:
            print('WARNING: Inconsistency between dihedral descriptions and dihedral type numbers found.')
            dih_types = ['']*header['dihedral types']

        del lines[:ndihedrals]

    opls_struct.dih_types = dih_types
    opls_struct.dih_list = dih_list


    return opls_struct


def update_from_lammps_dump(atoms, filename, check=True):
    atoms_dump = ase.io.lammpsrun.read_lammps_dump(filename)

    if len(atoms_dump) != len(atoms):
        raise RuntimeError('Structure in ' + filename +
                           ' has wrong length: %d != %d' %
                           (len(atoms_dump), len(atoms)))

    if check:
        for a, b in zip(atoms, atoms_dump):
            # check that the atom types match
            if not (a.tag + 1 == b.number):
                raise RuntimeError('Atoms index %d are of different '
                                   'type (%d != %d)' 
                                   % (a.index, a.tag + 1, b.number))

    atoms.set_cell(atoms_dump.get_cell())
    atoms.set_positions(atoms_dump.get_positions())
    if atoms_dump.get_velocities() is not None:
        atoms.set_velocities(atoms_dump.get_velocities())

    return atoms





# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

# Generic stuff may go here.

from matscipy.logger import screen
from .distributed_computation import BaseWorker, BaseResultManager
###

def has_parameter(name):
    """
    Test if a parameter has been provided in params.py.

    Parameters
    ----------
    name : str
        Name of the parameter.

    Returns
    -------
    value : bool
        Returns True if parameter exists.
    """
    import sys
    for x in ['.', '..']:
        if x not in sys.path:
            sys.path += [x]
    import params
    return name in params.__dict__


def parameter(name, default=None, logger=screen):
    """
    Read parameter from params.py control file.

    Parameters
    ----------
    name : str
        Name of the parameter.
    default : optional
        Default value. Will be returned if parameter is not present.

    Returns
    -------
    value
        Value of the parameter.
    """
    import sys
    for x in ['.', '..']:
        if x not in sys.path:
            sys.path += [x]
    import params
    try:
        value = params.__dict__[name]
        logger.pr('(user value)      {0} = {1}'.format(name, value))
    except KeyError:
        if default is not None:
            value = default
            logger.pr('(default value)   {0} = {1}'.format(name, value))
        else:
            raise
    return value

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014-2017) James Kermode, Warwick University
#                       Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

"""
Run calculation on a supercell of the atomic structure.
"""

import numpy as np

import ase
from ase.calculators.calculator import Calculator

###

class SupercellCalculator(Calculator):
    implemented_properties = ['energy', 'stress', 'forces']
    default_parameters = {}
    name = 'EAM'

    def __init__(self, calc, supercell):
        Calculator.__init__(self)
        self.calc = calc
        self.supercell = supercell

    def calculate(self, atoms, properties, system_changes):
        Calculator.calculate(self, atoms, properties, system_changes)

        atoms = self.atoms.copy()
        atoms.set_constraint(None)
        atoms *= self.supercell
        atoms.set_calculator(self.calc)

        energy = atoms.get_potential_energy()
        stress = atoms.get_stress()
        forces = atoms.get_forces()

        self.results = {'energy': energy/np.prod(self.supercell),
                        'stress': stress,
                        'forces': forces[:len(self.atoms)]}



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

"""
Embedded-atom method potential.
"""

import os

import numpy as np

import ase
from ase.calculators.calculator import Calculator

try:
    from scipy.interpolate import InterpolatedUnivariateSpline
except:
    InterpolatedUnivariateSpline = None

from scipy.sparse import bsr_matrix
from matscipy.calculators.eam.io import read_eam
from matscipy.neighbours import neighbour_list, first_neighbours, find_indices_of_reversed_pairs

###

def _make_splines(dx, y):
    if len(np.asarray(y).shape) > 1:
        return [_make_splines(dx, yy) for yy in y]
    else:
        return InterpolatedUnivariateSpline(np.arange(len(y))*dx, y)

def _make_derivative(x, n=1):
    if type(x) == list:
        return [_make_derivative(xx, n=n) for xx in x]
    else:
        return x.derivative(n=n)


###

class EAM(Calculator):
    implemented_properties = ['energy', 'stress', 'forces']
    default_parameters = {}
    name = 'EAM'
       
    def __init__(self, fn=None, atomic_numbers=None, F=None, f=None, rep=None,
                 cutoff=None, kind='eam/alloy'):
        Calculator.__init__(self)
        if fn is not None:
            source, parameters, F, f, rep = read_eam(fn, kind=kind)
            self._db_atomic_numbers = parameters.atomic_numbers
            self._db_cutoff = parameters.cutoff
            dr = parameters.distance_grid_spacing
            dF = parameters.density_grid_spacing

            # Create spline interpolation
            self.F = _make_splines(dF, F)
            self.f = _make_splines(dr, f)
            self.rep = _make_splines(dr, rep)
        else:
            self._db_atomic_numbers = atomic_numbers
            self.F = F
            self.f = f
            self.rep = rep
            self._db_cutoff = cutoff

        self.atnum_to_index = -np.ones(np.max(self._db_atomic_numbers)+1, dtype=int)
        self.atnum_to_index[self._db_atomic_numbers] = \
            np.arange(len(self._db_atomic_numbers))

        # Derivative of spline interpolation
        self.dF = _make_derivative(self.F)
        self.df = _make_derivative(self.f)
        self.drep = _make_derivative(self.rep)

        # Second derivative of spline interpolation
        self.ddF = _make_derivative(self.F, n=2)
        self.ddf = _make_derivative(self.f, n=2)
        self.ddrep = _make_derivative(self.rep, n=2)

    def energy_virial_and_forces(self, atomic_numbers_i, i_n, j_n, dr_nc, abs_dr_n):
        """
        Compute the potential energy, the virial and the forces.

        Parameters
        ----------
        atomic_numbers_i : array_like
            Atomic number for each atom in the system
        i_n, j_n : array_like
            Neighbor pairs
        dr_nc : array_like
            Distance vectors between neighbors
        abd_dr_n : array_like
            Length of distance vectors between neighbors

        Returns
        -------
        epot : float
            Potential energy
        virial_v : array
            Virial
        forces_ic : array
            Forces acting on each atom
        """
        nat = len(atomic_numbers_i)
        atnums_in_system = set(atomic_numbers_i)
        for atnum in atnums_in_system:
            if atnum not in self._db_atomic_numbers:
                raise RuntimeError('Element with atomic number {} found, but '
                                   'this atomic number has no EAM '
                                   'parametrization'.format(atnum))

        # Density
        f_n = np.zeros_like(abs_dr_n)
        df_n = np.zeros_like(abs_dr_n)
        for atidx1, atnum1 in enumerate(self._db_atomic_numbers):
            f1 = self.f[atidx1]
            df1 = self.df[atidx1]
            mask1 = atomic_numbers_i[j_n]==atnum1
            if mask1.sum() > 0:
                if type(f1) == list:
                    for atidx2, atnum2 in enumerate(self._db_atomic_numbers):
                        f = f1[atidx2]
                        df = df1[atidx2]
                        mask = np.logical_and(mask1, atomic_numbers_i[i_n]==atnum2)
                        if mask.sum() > 0:
                            f_n[mask] = f(abs_dr_n[mask])
                            df_n[mask] = df(abs_dr_n[mask])
                else:
                    f_n[mask1] = f1(abs_dr_n[mask1])
                    df_n[mask1] = df1(abs_dr_n[mask1])

        density_i = np.bincount(i_n, weights=f_n, minlength=nat)

        # Repulsion
        rep_n = np.zeros_like(abs_dr_n)
        drep_n = np.zeros_like(abs_dr_n)
        for atidx1, atnum1 in enumerate(self._db_atomic_numbers):
            rep1 = self.rep[atidx1]
            drep1 = self.drep[atidx1]
            mask1 = atomic_numbers_i[i_n]==atnum1
            if mask1.sum() > 0:
                for atidx2, atnum2 in enumerate(self._db_atomic_numbers):
                    rep = rep1[atidx2]
                    drep = drep1[atidx2]
                    mask = np.logical_and(mask1, atomic_numbers_i[j_n]==atnum2)
                    if mask.sum() > 0:
                        r = rep(abs_dr_n[mask])/abs_dr_n[mask]
                        rep_n[mask] = r
                        drep_n[mask] = (drep(abs_dr_n[mask])-r)/abs_dr_n[mask]

        # Energy
        epot = 0.5*np.sum(rep_n)
        demb_i = np.zeros(nat)
        for atidx, atnum in enumerate(self._db_atomic_numbers):
            F = self.F[atidx]
            dF = self.dF[atidx]
            mask = atomic_numbers_i==atnum
            if mask.sum() > 0:
                epot += np.sum(F(density_i[mask]))
                demb_i[mask] += dF(density_i[mask])

        # Forces
        reverse = find_indices_of_reversed_pairs(i_n, j_n, abs_dr_n)
        df_i_n = np.take(df_n, reverse)
        df_nc = -0.5*((demb_i[i_n]*df_n+demb_i[j_n]*df_i_n)+drep_n).reshape(-1,1)*dr_nc/abs_dr_n.reshape(-1,1)

        # Sum for each atom
        fx_i = np.bincount(j_n, weights=df_nc[:,0], minlength=nat) - \
            np.bincount(i_n, weights=df_nc[:,0], minlength=nat)
        fy_i = np.bincount(j_n, weights=df_nc[:,1], minlength=nat) - \
            np.bincount(i_n, weights=df_nc[:,1], minlength=nat)
        fz_i = np.bincount(j_n, weights=df_nc[:,2], minlength=nat) - \
            np.bincount(i_n, weights=df_nc[:,2], minlength=nat)

        # Virial
        virial_v = -np.array([dr_nc[:,0]*df_nc[:,0],               # xx
                              dr_nc[:,1]*df_nc[:,1],               # yy
                              dr_nc[:,2]*df_nc[:,2],               # zz
                              dr_nc[:,1]*df_nc[:,2],               # yz
                              dr_nc[:,0]*df_nc[:,2],               # xz
                              dr_nc[:,0]*df_nc[:,1]]).sum(axis=1)  # xy

        return epot, virial_v, np.transpose([fx_i, fy_i, fz_i])

    def calculate(self, atoms, properties, system_changes):
        Calculator.calculate(self, atoms, properties, system_changes)

        i_n, j_n, dr_nc, abs_dr_n = neighbour_list('ijDd', self.atoms,
                                                   self._db_cutoff)

        epot, virial_v, forces_ic = self.energy_virial_and_forces(self.atoms.numbers, i_n, j_n, dr_nc, abs_dr_n)

        self.results = {'energy': epot, 'free_energy': epot,
                        'stress': virial_v/self.atoms.get_volume(),
                        'forces': forces_ic}

    def calculate_hessian_matrix(self, atoms, divide_by_masses=False):
        r"""Compute the Hessian matrix

        The Hessian matrix is the matrix of second derivatives 
        of the potential energy :math:`\mathcal{V}_\mathrm{int}` 
        with respect to coordinates, i.e.\

        .. math:: 
        
            \frac{\partial^2 \mathcal{V}_\mathrm{int}}
                 {\partial r_{\nu{}i}\partial r_{\mu{}j}},

        where the indices :math:`\mu` and :math:`\nu` refer to atoms and
        the indices :math:`i` and :math:`j` refer to the components of the
        position vector :math:`r_\nu` along the three spatial directions.

        The Hessian matrix has contributions from the pair potential
        and the embedding energy, 

        .. math::

            \frac{\partial^2 \mathcal{V}_\mathrm{int}}{\partial r_{\nu{}i}\partial r_{\mu{}j}} = 
            \frac{\partial^2 \mathcal{V}_\mathrm{pair}}{ \partial r_{\nu i} \partial r_{\mu j}} +
            \frac{\partial^2 \mathcal{V}_\mathrm{embed}}{\partial r_{\nu i} \partial r_{\mu j}}. 	


        The contribution from the pair potential is

        .. math::

            \frac{\partial^2 \mathcal{V}_\mathrm{pair}}{ \partial r_{\nu i} \partial r_{\mu j}} &= 
            -\phi_{\nu\mu}'' \left(
            \frac{r_{\nu\mu i}}{r_{\nu\mu}} 
            \frac{r_{\nu\mu j}}{r_{\nu\mu}} 
            \right)
            -\frac{\phi_{\nu\mu}'}{r_{\nu\mu}}\left(
            \delta_{ij}-
            \frac{r_{\nu\mu i}}{r_{\nu\mu}} 
            \frac{r_{\nu\mu j}}{r_{\nu\mu}} 
            \right) \\ 
            &+\delta_{\nu\mu}\sum_{\gamma\neq\nu}^{N}
            \phi_{\nu\gamma}'' \left(
            \frac{r_{\nu\gamma i}}{r_{\nu\gamma}} 
            \frac{r_{\nu\gamma j}}{r_{\nu\gamma}} 
            \right)
            +\delta_{\nu\mu}\sum_{\gamma\neq\nu}^{N}\frac{\phi_{\nu\gamma}'}{r_{\nu\gamma}}\left(
            \delta_{ij}-
            \frac{r_{\nu\gamma i}}{r_{\nu\gamma}} 
            \frac{r_{\nu\gamma j}}{r_{\nu\gamma}} 
            \right).

        The contribution of the embedding energy to the Hessian matrix is a sum of eight terms,
        
        .. math::

            \frac{\mathcal{V}_\mathrm{embed}}{\partial r_{\mu j} \partial r_{\nu i}} 
            	&= T_1 + T_2 + T_3 + T_4 + T_5 + T_6 + T_7 + T_8 \\ 
            T_1 &= 
            \delta_{\nu\mu}U_\nu''
            \sum_{\gamma\neq\nu}^{N}g_{\nu\gamma}'\frac{r_{\nu\gamma i}}{r_{\nu\gamma}}
            \sum_{\gamma\neq\nu}^{N}g_{\nu\gamma}'\frac{r_{\nu\gamma j}}{r_{\nu\gamma}} \\
            T_2 &= 
            -u_\nu''g_{\nu\mu}' \frac{r_{\nu\mu j}}{r_{\nu\mu}} \sum_{\gamma\neq\nu}^{N} 
            G_{\nu\gamma}' \frac{r_{\nu\gamma i}}{r_{\nu\gamma}} \\
            T_3 &=
            +u_\mu''g_{\mu\nu}' \frac{r_{\nu\mu i}}{r_{\nu\mu}} \sum_{\gamma\neq\mu}^{N} 
            G_{\mu\gamma}' \frac{r_{\mu\gamma j}}{r_{\mu\gamma}} \\
            T_4 &= -\left(u_\mu'g_{\mu\nu}'' + u_\nu'g_{\nu\mu}''\right)
            \left(
            \frac{r_{\nu\mu i}}{r_{\nu\mu}} 
            \frac{r_{\nu\mu j}}{r_{\nu\mu}}
            \right)\\
            T_5 &= \delta_{\nu\mu} \sum_{\gamma\neq\nu}^{N}
            \left(U_\gamma'g_{\gamma\nu}'' + U_\nu'g_{\nu\gamma}''\right)
            \left(
            \frac{r_{\nu\gamma i}}{r_{\nu\gamma}}
            \frac{r_{\nu\gamma j}}{r_{\nu\gamma}}
            \right) \\
            T_6 &= -\left(U_\mu'g_{\mu\nu}' + U_\nu'g_{\nu\mu}'\right) \frac{1}{r_{\nu\mu}}
            \left(
            \delta_{ij}- 
            \frac{r_{\nu\mu i}}{r_{\nu\mu}} 
            \frac{r_{\nu\mu j}}{r_{\nu\mu}}
            \right) \\
            T_7 &= \delta_{\nu\mu} \sum_{\gamma\neq\nu}^{N}
            \left(U_\gamma'g_{\gamma\nu}' + U_\nu'g_{\nu\gamma}'\right) \frac{1}{r_{\nu\gamma}}
            \left(\delta_{ij}-
            \frac{r_{\nu\gamma i}}{r_{\nu\gamma}} 
            \frac{r_{\nu\gamma j}}{r_{\nu\gamma}}
            \right) \\
            T_8 &= \sum_{\substack{\gamma\neq\nu \\ \gamma \neq \mu}}^{N}
            U_\gamma'' g_{\gamma\nu}'g_{\gamma\mu}' 
            \frac{r_{\gamma\nu i}}{r_{\gamma\nu}}
            \frac{r_{\gamma\mu j}}{r_{\gamma\mu}} 


        Parameters
        ----------
        atoms : ase.Atoms
        divide_by_masses : bool
            Divide block :math:`\nu\mu` by :math:`m_\num_\mu` to obtain the dynamical matrix

        Returns
        -------
        D : numpy.matrix
            Block Sparse Row matrix with the nonzero blocks

        Notes
        -----
        Notation:
         * :math:`N` Number of atoms 
         * :math:`\mathcal{V}_\mathrm{int}`  Total potential energy 
         * :math:`\mathcal{V}_\mathrm{pair}` Pair potential 
         * :math:`\mathcal{V}_\mathrm{embed}` Embedding energy 
         * :math:`r_{\nu{}i}`  Component :math:`i` of the position vector of atom :math:`\nu` 
         * :math:`r_{\nu\mu{}i} = r_{\mu{}i}-r_{\nu{}i}` 
         * :math:`r_{\nu\mu{}}` Norm of :math:`r_{\nu\mu{}i}`, i.e.\ :math:`\left(r_{\nu\mu{}1}^2+r_{\nu\mu{}2}^2+r_{\nu\mu{}3}^2\right)^{1/2}`
         * :math:`\phi_{\nu\mu}(r_{\nu\mu{}})` Pair potential energy of atoms :math:`\nu` and :math:`\mu` 
         * :math:`\rho_nu` Total electron density of atom :math:`\nu`  
         * :math:`U_\nu(\rho_nu)` Embedding energy of atom :math:`\nu` 
         * :math:`g_{\delta}\left(r_{\gamma\delta}\right) \equiv g_{\gamma\delta}` Contribution from atom :math:`\delta` to :math:`\rho_\gamma`
         * :math:`m_\nu` mass of atom :math:`\nu`
        """

        nat = len(atoms)
        atnums = atoms.numbers

        atnums_in_system = set(atnums)
        for atnum in atnums_in_system:
            if atnum not in atnums:
                raise RuntimeError('Element with atomic number {} found, but '
                                   'this atomic number has no EAM '
                                   'parametrization'.format(atnum))

        # i_n: index of the central atom
        # j_n: index of the neighbor atom
        # dr_nc: distance vector between the two
        # abs_dr_n: norm of distance vector
        # Variable name ending with _n indicate arrays that contain
        # one element for each pair in the neighbor list. Names ending
        # with _i indicate arrays containing one element for each atom.
        i_n, j_n, dr_nc, abs_dr_n = neighbour_list('ijDd', atoms,
                                                   self._db_cutoff)
        first_i = first_neighbours(nat, i_n)
        # Make sure that the neighborlist does not contain the same pairs twice.
        # Reoccuring entries may be due to small system size. In this case, the
        # Hessian matrix will not be symmetric.
        unique_pairs = set((i, j) for (i, j) in zip(i_n, j_n))
        if len(unique_pairs) != len(i_n):
            raise ValueError("neighborlist contains some pairs more than once") 
        assert(np.all(i_n != j_n))

        if divide_by_masses: 
            masses_i = atoms.get_masses().reshape(-1, 1, 1)
            geom_mean_mass_n = np.sqrt(
                np.take(masses_i, i_n) * np.take(masses_i, j_n)
            ).reshape(-1, 1, 1)

        # Calculate the derivatives of the pair energy
        drep_n = np.zeros_like(abs_dr_n)  # first derivative
        ddrep_n = np.zeros_like(abs_dr_n) # second derivative
        for atidx1, atnum1 in enumerate(self._db_atomic_numbers):
            rep1 = self.rep[atidx1]
            drep1 = self.drep[atidx1]
            ddrep1 = self.ddrep[atidx1]
            mask1 = atnums[i_n]==atnum1
            if mask1.sum() > 0:
                for atidx2, atnum2 in enumerate(self._db_atomic_numbers):
                    rep = rep1[atidx2]
                    drep = drep1[atidx2]
                    ddrep = ddrep1[atidx2]
                    mask = np.logical_and(mask1, atnums[j_n]==atnum2)
                    if mask.sum() > 0:
                        r = rep(abs_dr_n[mask])/abs_dr_n[mask]
                        drep_n[mask] = (drep(abs_dr_n[mask])-r) / abs_dr_n[mask]
                        ddrep_n[mask] = (ddrep(abs_dr_n[mask]) - 2.0 * drep_n[mask]) / abs_dr_n[mask]

        # Calculate the total electron density at each atom, and the 
        # derivatives of pairwise contributions
        f_n = np.zeros_like(abs_dr_n)
        df_n = np.zeros_like(abs_dr_n)  # first derivative
        ddf_n = np.zeros_like(abs_dr_n) # second derivative
        for atidx1, atnum1 in enumerate(self._db_atomic_numbers):
            f1 = self.f[atidx1]     
            df1 = self.df[atidx1]   
            ddf1 = self.ddf[atidx1] 
            mask1 = atnums[j_n]==atnum1
            if mask1.sum() > 0:
                if type(f1) == list:
                    for atidx2, atnum2 in enumerate(self._db_atomic_numbers):
                        f = f1[atidx2]
                        df = df1[atidx2]
                        ddf = ddf1[atidx2]
                        mask = np.logical_and(mask1, atnums[i_n]==atnum2)
                        if mask.sum() > 0:
                            f_n[mask] = f(abs_dr_n[mask])
                            df_n[mask] = df(abs_dr_n[mask])
                            ddf_n[mask] = ddf(abs_dr_n[mask])
                else:
                    f_n[mask1] = f1(abs_dr_n[mask1])
                    df_n[mask1] = df1(abs_dr_n[mask1])
                    ddf_n[mask1] = ddf1(abs_dr_n[mask1])
        # Accumulate density contributions
        density_i = np.bincount(i_n, weights=f_n, minlength=nat)

        # Calculate the derivatives of the embedding energy
        demb_i = np.zeros(nat)   # first derivative
        ddemb_i = np.zeros(nat)  # second derivative
        for atidx, atnum in enumerate(self._db_atomic_numbers):
            F = self.F[atidx]
            dF = self.dF[atidx]
            ddF = self.ddF[atidx]
            mask = atnums==atnum
            if mask.sum() > 0:
                demb_i[mask] += dF(density_i[mask])
                ddemb_i[mask] += ddF(density_i[mask])

        symmetry_check = False # check symmetry of individual terms?
        #------------------------------------------------------------------------ 
        # Calculate pair contribution to the Hessian matrix
        #------------------------------------------------------------------------ 
        e_nc = (dr_nc.T / abs_dr_n).T # normalized distance vectors r_i^{\mu\nu}
        outer_1 = e_nc.reshape(-1, 3, 1) * e_nc.reshape(-1, 1, 3)
        outer_2 = np.eye(3, dtype=e_nc.dtype) - outer_1
        D_ncc = -(ddrep_n * outer_1.T).T
        D_ncc += -(drep_n / abs_dr_n * outer_2.T).T
        if divide_by_masses:
            D = bsr_matrix(
                (D_ncc / geom_mean_mass_n, j_n, first_i), 
                shape=(3*nat, 3*nat)
            )
        else:
            D = bsr_matrix((D_ncc, j_n, first_i), shape=(3*nat, 3*nat))
        Ddiag = np.empty((nat, 3, 3))
        for x in range(3):
            for y in range(3):
                Ddiag[:, x, y] = -np.bincount(i_n, weights=D_ncc[:, x, y]) # summation
        if divide_by_masses:
            Ddiag /= masses_i
        # put 3x3 blocks on diagonal (Kronecker Delta delta_{\mu\nu})
        D += bsr_matrix((Ddiag, np.arange(nat), np.arange(nat+1)), shape=(3*nat, 3*nat))

        #------------------------------------------------------------------------ 
        # Calculate contribution of embedding term
        #------------------------------------------------------------------------ 
        # For each pair in the neighborlist, create arrays which store the 
        # derivatives of the embedding energy of the corresponding atoms.
        demb_i_n = np.take(demb_i, i_n)
        demb_j_n = np.take(demb_i, j_n)
        ddemb_i_n = np.take(ddemb_i, i_n)
        ddemb_j_n = np.take(ddemb_i, j_n)

        # Let r be an index into the neighbor list. df_n[r] contains the the
        # contribution from atom j_n[r] to the derivative of the electron
        # density of atom i_n[r]. We additionally need the contribution of
        # i_n[r] to the derivative of j_n[r]. This value is also in df_n,
        # but at a different position. reverse[r] gives the new index s
        # where we find this value. The same indexing applies to ddf_n.
        reverse = find_indices_of_reversed_pairs(i_n, j_n, abs_dr_n)
        df_i_n = np.take(df_n, reverse)
        ddf_i_n = np.take(ddf_n, reverse)
        #we already have ddf_j_n = ddf_n 

        # Term 1: 
        # \delta_{\nu\mu}U_\nu''
        # \sum_{\gamma\neq\nu}^{\natoms}g_{\nu\gamma}'\frac{r_{\nu\gamma i}}{r_{\nu\gamma}} 
        # \sum_{\gamma\neq\nu}^{\natoms}g_{\nu\gamma}'\frac{r_{\nu\gamma j}}{r_{\nu\gamma}} 
        # Likely zero in equilibrium because the sum is zero (appears in the force vector)
        df_n_e_nc_outer_product = (df_n * e_nc.T).T
        df_n_e_nc_i = np.empty((nat, 3), dtype=df_n.dtype)
        for x in range(3):
            df_n_e_nc_i[:, x] = np.bincount(i_n, weights=df_n_e_nc_outer_product[:, x], minlength=nat)
        term_1_ncc = ((ddemb_i * df_n_e_nc_i.T).T).reshape(-1,3,1) * df_n_e_nc_i.reshape(-1,1,3)
        if divide_by_masses:
            term_1_ncc /= masses_i
        term_1 = bsr_matrix((term_1_ncc, np.arange(nat), np.arange(nat+1)), shape=(3*nat, 3*nat)) 
        D += term_1
        if symmetry_check: 
            print("check term 1", np.linalg.norm(term_1.todense() - term_1.todense().T))

        # Term 2: 
        # -u_\nu''g_{\nu\mu}' \frac{r_{\nu\mu j}}{r_{\nu\mu}} \sum_{\gamma\neq\nu}^{\natoms} 
        #      g_{\nu\gamma}' \frac{r_{\nu\gamma i}}{r_{\nu\gamma}}
        # Likely zero in equilibrium because the sum is zero (appears in the force vector)
        df_n_e_nc_j_n = np.take(df_n_e_nc_i, j_n, axis=0)
        term_2_ncc = ((ddemb_j_n * df_i_n * e_nc.T).T).reshape(-1,3,1) * df_n_e_nc_j_n.reshape(-1,1,3)
        if divide_by_masses:
            term_2_ncc /= geom_mean_mass_n
        term_2 = bsr_matrix((term_2_ncc, j_n, first_i), shape=(3*nat, 3*nat))
        D += term_2 
        if symmetry_check: 
            print("check term 2", np.linalg.norm(term_2.todense() - term_2.todense().T))

        # Term 3: 
        # +u_\mu''g_{\mu\nu}' \frac{r_{\nu\mu i}}{r_{\nu\mu}} \sum_{\gamma\neq\mu}^{\natoms} 
        #      g_{\mu\gamma}' \frac{r_{\mu\gamma j}}{r_{\mu\gamma}} 
        # Likely zero in equilibrium because the sum is zero (appears in the force vector)
        df_n_e_nc_i_n = np.take(df_n_e_nc_i, i_n, axis=0)
        term_3_ncc = -((ddemb_i_n * df_n * df_n_e_nc_i_n.T).T).reshape(-1,3,1) * e_nc.reshape(-1,1,3)
        if divide_by_masses:
            term_3_ncc /= geom_mean_mass_n
        term_3 = bsr_matrix((term_3_ncc, j_n, first_i), shape=(3*nat, 3*nat))
        D += term_3
        if symmetry_check: 
            print("check term 3", np.linalg.norm(term_3.todense() - term_3.todense().T))

        # Term 4:
        # -\left(u_\mu'g_{\mu\nu}'' + u_\nu'g_{\nu\mu}''\right)
        # \left(
        # \frac{r_{\nu\mu i}}{r_{\nu\mu}} 
        # \frac{r_{\nu\mu j}}{r_{\nu\mu}}
        # \right)
        tmp_1 = -((demb_j_n * ddf_i_n + demb_i_n * ddf_n) * outer_1.T).T 
        # We don't immediately add term 4 to the matrix, because it would have 
        # to be normalized by the masses if divide_by_masses is true. However,
        # for construction of term 5, we need term 4 without normalization

        # Term 5:
        # \delta_{\nu\mu} \sum_{\gamma\neq\nu}^{\natoms}
        #\left(U_\gamma'g_{\gamma\nu}'' + U_\nu'g_{\nu\gamma}''\right)
        #\left(
        #\frac{r_{\nu\gamma i}}{r_{\nu\gamma}}
        #\frac{r_{\nu\gamma j}}{r_{\nu\gamma}}
        #\right)
        tmp_1_summed = np.empty((nat, 3, 3), dtype=tmp_1.dtype)
        for x in range(3):
            for y in range(3):
                tmp_1_summed[:, x, y] = -np.bincount(i_n, weights=tmp_1[:, x, y]) 
        if divide_by_masses:
            tmp_1_summed /= masses_i
        term_5 = bsr_matrix((tmp_1_summed, np.arange(nat), np.arange(nat+1)), shape=(3*nat, 3*nat))
        D += term_5
        if symmetry_check: 
            print("check term 5", np.linalg.norm(term_5.todense() - term_5.todense().T))
        if divide_by_masses:
            tmp_1 /= geom_mean_mass_n
        term_4 = bsr_matrix((tmp_1, j_n, first_i), shape=(3*nat, 3*nat))
        D += term_4 
        if symmetry_check: 
            print("check term 4", np.linalg.norm(term_4.todense() - term_4.todense().T))

        # Term 6:
        # -\left(U_\mu'g_{\mu\nu}' + U_\nu'g_{\nu\mu}'\right) \frac{1}{r_{\nu\mu}}
        #\left(
        #\delta_{ij}- 
        #\frac{r_{\nu\mu i}}{r_{\nu\mu}} 
        #\frac{r_{\nu\mu j}}{r_{\nu\mu}}
        #\right)
        # Like term 4, which was needed to construct term 5, we don't add 
        # term 6 immediately, because it is needed for construction of term 7
        tmp_2 = -((demb_j_n * df_i_n + demb_i_n * df_n) / abs_dr_n * outer_2.T).T

        # Term 7:
        # \delta_{\nu\mu} \sum_{\gamma\neq\nu}^{\natoms}
        #\left(U_\gamma'g_{\gamma\nu}' + U_\nu'g_{\nu\gamma}'\right) \frac{1}{r_{\nu\gamma}}
        #\left(\delta_{ij}-
        #\frac{r_{\nu\gamma i}}{r_{\nu\gamma}}
        #\frac{r_{\nu\gamma j}}{r_{\nu\gamma}}
        #\right)
        tmp_2_summed = np.empty((nat, 3, 3), dtype=tmp_2.dtype)
        for x in range(3):
            for y in range(3):
                tmp_2_summed[:, x, y] = -np.bincount(i_n, weights=tmp_2[:, x, y]) 
        if divide_by_masses:
            tmp_2_summed /= masses_i
        term_7 = bsr_matrix((tmp_2_summed, np.arange(nat), np.arange(nat+1)), shape=(3*nat, 3*nat))
        D += term_7 
        if symmetry_check: 
            print("check term 7", np.linalg.norm(term_7.todense() - term_7.todense().T))
        if divide_by_masses:
            tmp_2 /= geom_mean_mass_n
        term_6 = bsr_matrix((tmp_2, j_n, first_i), shape=(3*nat, 3*nat))
        D += term_6 
        if symmetry_check: 
            print("check term 6", np.linalg.norm(term_6.todense() - term_6.todense().T))

        #Term 8: 
        #  \sum_{\substack{\gamma\neq\nu \\ \gamma \neq \mu}}^{\natoms}
        #U_\gamma'' g_{\gamma\nu}'g_{\gamma\mu}' 
        #\frac{r_{\gamma\nu i}}{r_{\gamma\nu}}
        #\frac{r_{\gamma\mu j}}{r_{\gamma\mu}} 
        #-----------------------------------------------------------------------
        # This term requires knowledge of common neighbors of pairs of atoms.
        # Construction of a common neighbor list in Python is likely a
        # performance bottleneck; it should be implemented in C++ instead.
        #-----------------------------------------------------------------------

        # For each (i1, j2)-pair in the neighbor list, find all other
        # pairs (i2, j) which share the same j. This includes (i1, j)
        # itself. In this way, create a list with n blocks of rows, where
        # n is the length of the neighbor list. All rows in a block have
        # the same j. Each row corresponds to one triplet i1, j, i2.
        # The number of rows in the block is equal to the total number
        # of neighbors of j. This is the common neighbor list (cnl).
        # Besides the block number and i1, j, i2, the cnl stores the
        # index into the neighbor list where the pair i2-j can be found.
        j_order = np.argsort(j_n)
        i_n_2 = i_n[j_order]
        j_n_2 = j_n[j_order]
        first_j = first_neighbours(nat, j_n_2)
        num_rows_per_j = first_j[j_n+1] - first_j[j_n]
        total_num_rows = np.sum(num_rows_per_j)
        # The common neighbor information could be stored as
        # a 2D array. However, multiple 1D arrays are likely
        # better for performance (fewer cache misses later).
        cnl_block_number = np.empty(total_num_rows, dtype=i_n.dtype)
        cnl_j1 = np.empty(total_num_rows, dtype=i_n.dtype)
        cnl_j_order = np.empty(total_num_rows, dtype=i_n.dtype)
        cnl_i1_i2 = np.empty((total_num_rows, 2), dtype=i_n.dtype)
        block_start = np.r_[0, np.cumsum(num_rows_per_j)]
        slice_for_j1 = {j1: slice(first_j[j1], first_j[j1+1]) for j1 in np.arange(nat)}
        for block_number, (i1, j1) in enumerate(zip(i_n, j_n)):
            slice1 = slice(block_start[block_number], block_start[block_number+1])
            slice2 = slice_for_j1[j1] 
            cnl_block_number[slice1] = block_number
            cnl_j1[slice1] = j1 
            cnl_j_order[slice1] = j_order[slice2]
            cnl_i1_i2[slice1, 0] = i1
            cnl_i1_i2[slice1, 1] = i_n_2[slice2]
        # Determine bins for accumulation by bincount
        unique_pairs_i1_i2, bincount_bins = np.unique(cnl_i1_i2, axis=0, return_inverse=True)
        e_nu_cnl = np.take(e_nc, cnl_block_number, axis=0)
        e_mu_cnl = np.take(e_nc, cnl_j_order, axis=0)
        ddemb_cnl = np.take(ddemb_i, cnl_j1)
        df_nu_cnl = np.take(df_i_n, cnl_block_number)
        df_mu_cnl = np.take(df_i_n, cnl_j_order)
        tmp_3 = ((ddemb_cnl * df_nu_cnl * df_mu_cnl * e_nu_cnl.T).T).reshape(-1, 3, 1) * e_mu_cnl.reshape(-1, 1, 3)
        tmp_3_summed = np.empty((unique_pairs_i1_i2.shape[0], 3, 3), dtype=e_nc.dtype)
        for x in range(3):
            for y in range(3):
                tmp_3_summed[:, x, y] = np.bincount(
                        bincount_bins, weights=tmp_3[:, x, y], 
                        minlength=unique_pairs_i1_i2.shape[0]
                ) 
        if divide_by_masses:
            geom_mean_mass_i1_i2 = np.sqrt(
                np.take(masses_i, unique_pairs_i1_i2[:, 0]) * np.take(masses_i, unique_pairs_i1_i2[:, 1])
            )
            tmp_3_summed /= geom_mean_mass_i1_i2[:, np.newaxis, np.newaxis]
        index_ptr = first_neighbours(nat, unique_pairs_i1_i2[:, 0])
        term_8 = bsr_matrix((tmp_3_summed, unique_pairs_i1_i2[:, 1], index_ptr), shape=(3*nat, 3*nat))
        if symmetry_check:
            print("check term 8", np.linalg.norm(term_8.todense() - term_8.todense().T))
        D += term_8
        return D

    @property
    def cutoff(self):
        return self._db_cutoff



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

import re

import numpy as np

###

def savetbl(fn, **kwargs):
    """
    Save tabulated data and write column header strings.

    Example:
        savetbl('file.dat', time=time, strain=strain, energy=energy)

    Parameters
    ----------
    fn : str
        Name of file to write to.
    kwargs : dict
        Keyword argument pass data and column header names.
    """
    sorted_kwargs = sorted(kwargs.items(), key=lambda t: t[0])
    header = ''
    for i, (x, y) in enumerate(sorted_kwargs):
        header = '{0} {1}:{2}'.format(header, i+1, x)
    data = np.transpose([y for x, y in sorted_kwargs])
    np.savetxt(fn, data, header=header)


def loadtbl(fn, usecols=None, converters=None, fromfile=False, **kwargs):
    """
    Load tabulated data from column header strings.

    Example data file:
        # time strain energy
        1.0 0.01 283
        2.0 0.02 398
        ...

        strain, energy = loadtbl('file.dat', usecols=['strain', 'energy'])

    Parameters
    ----------
    fn : str
        Name of file to load.
    usecols : list of strings
        List of column names.
    fromfile : bool
        Use numpy.fromfile instead of numpy.loadtxt if set to True. Can be
        faster in some circumstances.

    Returns
    -------
    data : tuple of arrays
        Return tuple of array with data for each colume in usecols if
        usecols specified. For usecols=None, return dictionary with header
        keys and arrays as data entries.
    """
    f = open(fn)
    line = f.readline()
    column_labels = None
    while line.startswith('#'):
        line = line[1:].strip()
        column_labels = [s.strip() for s in re.split('[\s,]+', line)]
        line = f.readline()
    if column_labels is None:
        f.close()
        raise RuntimeError("No header found in file '{}'".format(fn))

    sep_i = [x.find(':') for x in column_labels]
    column_labels = [s[i+1:] if i >= 0 else s for s, i
                     in zip(column_labels, sep_i)]

    if fromfile:
        data = np.fromfile(f, sep=' ')
        f.close()
        data.shape = (-1, len(column_labels))
        if usecols is None:
            return dict((s, d) for s, d in zip(column_labels, data.T))
        else:
            return [data[:, column_labels.index(s)] for s in usecols]
    else:
        f.close()
        converters_i = None
        if converters is not None:
            converters_i = {column_labels.index(s): c for s, c
                            in converters.items()}
        if usecols is None:
            data = np.loadtxt(fn, converters=converters_i, unpack=True,
                              **kwargs)
            return dict((s, d) for s, d in zip(column_labels, data))
        else:
            column_i = [column_labels.index(s) for s in usecols]
            return np.loadtxt(fn, usecols=column_i, converters=converters_i,
                              unpack=True, **kwargs)



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

import itertools
import functools

import numpy as np
from numpy.linalg import norm, inv

def gcd(a, b):
    """Calculate the greatest common divisor of a and b"""
    while b:
        a, b = b, a%b
    return a

class MillerIndex(np.ndarray):
    """
    Representation of a three of four index Miller direction or plane

    A :class:`MillerIndex` can be constructed from vector or parsed from a string::

        x = MillerIndex('-211')
        y = MillerIndex('111', type='plane')
        z = x.cross(y)
        print x # prints "[-211]"
        print y # prints "(111)", note round brackets denoting a plane
        print z.latex()
        assert(angle_between(x,y) == pi/2.)
        assert(angle_between(y,z) == pi/2.)
        assert(angle_between(x,z) == pi/2.)
    """

    __array_priority__ = 101.0

    brackets = {'direction': '[]',
                'direction_family': '<>',
                'plane': '()',
                'plane_family': '{}'}

    all_brackets = list(itertools.chain(*brackets.values()))

    def __new__(cls, v=None, type='direction'):
        if isinstance(v, basestring):
            v = MillerIndex.parse(v)
        if len(v) == 3 or len(v) == 4:
            self = np.ndarray.__new__(cls, len(v))
            self[:] = v
        else:
            raise ValueError('%s input v should be of length 3 or 4' % cls.__name__)
        self.type = type
        self.simplify()
        return self

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.type = getattr(obj, 'type', 'direction')

    def __repr__(self):
        return ('%s(['+'%d'*len(self)+'])') % ((self.__class__.__name__,) + tuple(self))
        
    def __str__(self):
        bopen, bclose = MillerIndex.brackets[self.type]
        return (bopen+'%d'*len(self)+bclose) % tuple(self)

    def latex(self):
        """
        Format this :class:`MillerIndex` as a LaTeX string
        """
        s = '$'
        bopen, bclose = MillerIndex.brackets[self.type]
        s += bopen
        for component in self:
            if component < 0:
                s += r'\bar{%d}' % abs(component)
            else:
                s += '%d' % component
        s += bclose
        s += '$'
        return s

    @classmethod
    def parse(cls, s):
        r"""
        Parse a Miller index string

        Negative indices can be denoted by:
         1. leading minus sign, e.g. ``[11-2]``
         2. trailing ``b`` (for 'bar'), e.g. ``112b``
         3. LaTeX ``\bar{}``, e.g. ``[11\bar{2}]`` (which renders as :math:`[11\bar{2}]` in LaTeX)

        Leading or trailing brackets of various kinds are ignored.
        i.e. ``[001]``, ``{001}``, ``(001)``, ``[001]``, ``<001>``, ``001`` are all equivalent.

        Returns an array of components (i,j,k) or (h,k,i,l)
        """

        if not isinstance(s, str):
            raise TypeError("Can't parse from %r of type %r" % (s, type(s)))

        orig_s = s
        for (a, b) in [(r'\bar{','-')] + [(b,'') for b in MillerIndex.all_brackets]:
            s = s.replace(a, b)

        L = list(s)
        components = np.array([1,1,1,1]) # space for up to 4 elements
        i = 3 # parse backwards from end of string
        while L:
            if i < 0:
                raise ValueError('Cannot parse Miller index from string "%s", too many components found' % orig_s)
            c = L.pop()
            if c == '-':
                if i == 3:
                    raise ValueError('Miller index string "%s" cannot end with a minus sign' % orig_s)
                components[i+1] *= -1
            elif c == 'b':
                components[i] *= -1
            elif c in ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']:
                components[i] *= int(c)
                i -= 1
            else:
                raise ValueError('Unexpected character "%s" in miller index string "%s"' % (c, orig_s))
                
        if i == 0:
            return components[1:]
        elif i == -1:
            return components
        else:
            raise ValueError('Cannot parse Miller index from string %s, too few components found' % orig_s)
        
        self.simplify()

    def simplify(self):
        """
        Simplify by dividing through by greatest common denominator
        """
        d = abs(functools.reduce(gcd, self))
        self[:] /= d

    def simplified(self):
        copy = self.copy()
        copy.simplify()
        return copy

    def norm(self):
        return np.linalg.norm(self)

    def normalised(self):
        a = self.as3()
        return np.array(a, dtype=float)/a.norm()

    hat = normalised

    def cross(self, other):
        a = self.as3()
        b = MillerIndex(other).as3()
        return np.cross(a, b).view(MillerIndex).simplified()

    def cosine(self, other):
        other = MillerIndex(other)
        return np.dot(self.normalised(), other.normalised())

    def angle(self, other):
        return np.arccos(self.cosine(other))

    def as4(self):
        if len(self) == 4:
            return self
        else:
            h, k, l = self
            i = -(h+l)
            return MillerIndex((h,k,i,l))

    def as3(self):
        if len(self) == 3:
            return self
        else:
            h, k, i, l = self
            return MillerIndex((h, k, l))

    def plane_spacing(self, a):
        return a/self.as3().norm()

def MillerPlane(v):
   """Special case of :class:`MillerIndex` with ``type="plane"``"""
   return MillerIndex(v, 'plane')

def MillerDirection(v):
   """Special case of :class:`MillerIndex` with ``type="direction"`` (the default)"""
   return MillerIndex(v, 'direction')


def angle_between(a, b):
    """Angle between crystallographic directions between a=[ijk] and b=[lmn], in radians."""
    return MillerIndex(a).angle(b)


def make_unit_slab(unit_cell, axes):   
    """
    General purpose unit slab creation routine

    Only tested with cubic unit cells.

    Code translated from quippy.structures.unit_slab()
        https://github.com/libAtoms/QUIP/blob/public/src/libAtoms/Structures.f95

    Arguments
    ---------
        unit_cell : Atoms
            Atoms object containing primitive unit cell
        axes: 3x3 array
            Miller indices of desired slab, as columns

    Returns
    -------
        slab : Atoms
            Output slab, with axes aligned with x, y, z.
    """
    a1 = axes[:,0]
    a2 = axes[:,1]
    a3 = axes[:,2]
    rot = np.zeros((3,3))
    rot[0,:] = a1/norm(a1)
    rot[1,:] = a2/norm(a2)
    rot[2,:] = a3/norm(a3)
    
    pos = unit_cell.get_positions().T
    lattice = unit_cell.get_cell().T
    lattice = np.dot(rot, lattice)
    
    at = unit_cell.copy()
    at.set_positions(np.dot(rot, pos).T)
    at.set_cell(lattice.T)

    sup = at * (5,5,5)
    sup.positions[...] -= sup.positions.mean(axis=0)
    
    sup_lattice = np.zeros((3,3))
    for i in range(3):
        sup_lattice[:,i] = (axes[0,i]*lattice[:,0] + 
                            axes[1,i]*lattice[:,1] + 
                            axes[2,i]*lattice[:,2])
    
    sup.set_cell(sup_lattice.T, scale_atoms=False)
    
    # Form primitive cell by discarding atoms with 
    # lattice coordinates outside range [-0.5,0.5]
    d = [0.01,0.02,0.03]  # Small shift to avoid conincidental alignments
    i = 0
    g = inv(sup_lattice)
    sup_pos = sup.get_positions().T
    while True:
        t = np.dot(g, sup_pos[:, i] + d)
        if (t <= -0.5).any() | (t >= 0.5).any():
            del sup[i]
            sup_pos = sup.get_positions().T
            i -= 1 # Retest since we've removed an atom
        if i == len(sup)-1:
            break
        i += 1

    sup.set_scaled_positions(sup.get_scaled_positions())
    return sup    



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#                  Adrien Gola, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# ======================================================================


from __future__ import division, print_function

from collections import namedtuple

import numpy as np
try:
    from scipy import interpolate
except:
    print('Warning: No scipy')
    interpolate = False

import os

###

EAMParameters = namedtuple('EAMParameters', 'symbols atomic_numbers '
                           'atomic_masses lattice_constants crystal_structures '
                           'number_of_density_grid_points '
                           'number_of_distance_grid_points '
                           'density_grid_spacing distance_grid_spacing '
                           'cutoff')

###

def read_eam(eam_file,kind="eam/alloy"):
    """
    Read an eam alloy lammps format file and return the tabulated data and parameters
    http://lammps.sandia.gov/doc/pair_eam.html
    
    Parameters
    ----------
      eam_file : string
                      eam alloy file name 
      kind : string
             kind of EAM file to read (supported eam,eam/alloy,eam/fs)
    Returns
    -------
      source : string
          Source informations or comment line for the file header
      parameters : list of tuples
                [0] - array of str - atoms (ONLY FOR eam/alloy and eam/fs, EMPTY for eam)
                [1] - array of int - atomic numbers
                [2] - array of float -atomic masses
                [3] - array of float - equilibrium lattice parameter
                [4] - array of str - crystal structure
                [5] - int - number of data point for embedded function
                [6] - int - number of data point for density and pair functions
                [7] - float - step size for the embedded function
                [8] - float - step size for the density and pair functions
                [9] - float - cutoff of the potentials
      F : array_like
          contain the tabulated values of the embedded functions
          shape = (nb atoms, nb of data points)
      f : array_like
          contain the tabulated values of the density functions
          shape = (nb atoms, nb of data points)
      rep : array_like
          contain the tabulated values of pair potential
          shape = (nb atoms,nb atoms, nb of data points)
    """
    with open(eam_file, 'r') as file:
        eam = file.readlines()
    if kind=="eam":
        # reading first comment line as source for eam potential data
        source = eam[0].strip()
        # -- Parameters -- #
        atnumber = int(eam[1].split()[0])
        atmass = float(eam[1].split()[1])
        crystallatt = float(eam[1].split()[2])
        crystal = eam[1].split()[3]
        Nrho = int(eam[2].split()[0])       # Nrho (number of values for the embedding function F(rho))
        Nr = int(eam[2].split()[2])         # Nr (number of values for the effective charge function Z(r) and density function rho(r))
        drho = float(eam[2].split()[1])     # spacing in density space
        dr = float(eam[2].split()[3]) # spacing in distance space
        cutoff = float(eam[2].split()[4])
        parameters = EAMParameters(np.zeros(1),atnumber, atmass,crystallatt,crystal, Nrho,Nr, drho, dr, cutoff)
        # -- Tabulated data -- #
        data = np.loadtxt(eam_file, dtype="float", skiprows = 3).flatten()
        F = data[0:Nrho]
        f = data[Nrho:Nrho+Nr]
        rep = data[Nrho+Nr:2*Nr+Nrho]
        return source,parameters, F,f,rep
    elif kind=="eam/alloy":
        # reading 3 first comment lines as source for eam potential data
        source = eam[0].strip()+eam[1].strip()+eam[2].strip()
        # -- Parameters -- #
        atoms = eam[3].strip().split()[1:]
        nb_atoms = len(atoms)
        Nrho = int(eam[4].split()[0])       # Nrho (number of values for the embedding function F(rho))
        Nr = int(eam[4].split()[2])         # Nr (number of values for the effective charge function Z(r) and density function rho(r))
        drho = float(eam[4].split()[1])     # spacing in density space
        dr = float(eam[4].split()[3]) # spacing in distance space
        cutoff = float(eam[4].split()[4])
        atnumber,atmass,crystallatt,crystal = np.empty(nb_atoms,dtype=int),np.empty(nb_atoms),np.empty(nb_atoms),np.empty(nb_atoms).astype(np.str)
        for i in range(nb_atoms):
            # Fixme: The following lines assume that data occurs in blocks of
            # homogeneous width. This can break.
            l = len(eam[6].strip().split())
            row = int(5+i*((Nr+Nrho)/l+1))
            atnumber[i] = int(eam[row].split()[0])
            atmass[i] = float(eam[row].split()[1])
            crystallatt[i] = float(eam[row].split()[2])
            crystal[i] = str(eam[row].split()[3])
        parameters = EAMParameters(atoms,atnumber,atmass,crystallatt,crystal,Nrho,Nr,drho,dr,cutoff)
        # -- Tabulated data -- #
        F,f,rep,data = np.empty((nb_atoms,Nrho)),np.empty((nb_atoms,Nr)),np.empty((nb_atoms,nb_atoms,Nr)),np.empty(())
        eam = open(eam_file,'r')
        [eam.readline() for i in range(5)]
        for i in range(nb_atoms):
            eam.readline()
            data = np.append(data,np.fromfile(eam,count=Nrho+Nr, sep=' '))
        data = np.append(data,np.fromfile(eam,count=-1, sep=' '))
        data = data[1:]
        for i in range(nb_atoms):
            F[i,:] = data[i*(Nrho+Nr):Nrho+i*(Nrho+Nr)]
            f[i,:] = data[Nrho+i*(Nrho+Nr):Nrho+Nr+i*(Nrho+Nr)]
        interaction = 0
        for i in range(nb_atoms):
            for j in range(nb_atoms):
                if j < i :
                    rep[i,j,:] = data[nb_atoms*(Nrho+Nr)+interaction*Nr:nb_atoms*(Nrho+Nr)+interaction*Nr+Nr]
                    rep[j,i,:] = data[nb_atoms*(Nrho+Nr)+interaction*Nr:nb_atoms*(Nrho+Nr)+interaction*Nr+Nr]
                    interaction+=1
            rep[i,i,:] = data[nb_atoms*(Nrho+Nr)+interaction*Nr:nb_atoms*(Nrho+Nr)+interaction*Nr+Nr]
            interaction+=1
        return source,parameters, F,f,rep
    elif kind=="eam/fs":
        # reading 3 first comment lines as source for eam potential data
        source = eam[0].strip()+eam[1].strip()+eam[2].strip()
        # -- Parameters -- #
        atoms = eam[3].strip().split()[1:]
        nb_atoms = len(atoms)
        Nrho = int(eam[4].split()[0])       # Nrho (number of values for the embedding function F(rho))
        Nr = int(eam[4].split()[2])         # Nr (number of values for the effective charge function Z(r) and density function rho(r))
        drho = float(eam[4].split()[1])     # spacing in density space
        dr = float(eam[4].split()[3]) # spacing in distance space
        cutoff = float(eam[4].split()[4])
        atnumber,atmass,crystallatt,crystal = np.empty(nb_atoms,dtype=int),np.empty(nb_atoms),np.empty(nb_atoms),np.empty(nb_atoms).astype(np.str)
        for i in range(nb_atoms):
            # Fixme: The following lines assume that data occurs in blocks of
            # homogeneous width. This can break.
            l = len(eam[6].strip().split())
            row = int(5+i*((Nr*nb_atoms+Nrho)/l+1))
            atnumber[i] = int(eam[row].split()[0])
            atmass[i] = float(eam[row].split()[1])
            crystallatt[i] = float(eam[row].split()[2])
            crystal[i] = str(eam[row].split()[3])
        parameters = EAMParameters(atoms,atnumber,atmass,crystallatt,crystal,Nrho,Nr,drho,dr,cutoff)
        # -- Tabulated data -- #
        F,f,rep,data = np.empty((nb_atoms,Nrho)),np.empty((nb_atoms,nb_atoms,Nr)),np.empty((nb_atoms,nb_atoms,Nr)),np.empty(())
        eam = open(eam_file,'r')
        [eam.readline() for i in range(5)]
        for i in range(nb_atoms):
            eam.readline()
            data = np.append(data,np.fromfile(eam,count=Nrho+Nr*nb_atoms, sep=' '))
        data = np.append(data,np.fromfile(eam,count=-1, sep=' '))
        data = data[1:]
        for i in range(nb_atoms):
            F[i,:] = data[i*(Nrho+Nr*nb_atoms):Nrho+i*(Nrho+Nr*nb_atoms)]
            #f[i,:] = data[Nrho+i*(Nrho+Nr):Nrho+Nr+i*(Nrho+Nr)]
        interaction = 0
        for i in range(nb_atoms):
            for j in range(nb_atoms):
                f[i,j,:] = data[Nrho+j*Nr+i*(Nrho+Nr*nb_atoms):Nr+Nrho+j*Nr+i*(Nrho+Nr*nb_atoms)]
                if j < i :
                    rep[i,j,:] = data[nb_atoms*(Nrho+Nr*nb_atoms)+interaction*Nr:nb_atoms*(Nrho+Nr*nb_atoms)+interaction*Nr+Nr]
                    rep[j,i,:] = data[nb_atoms*(Nrho+Nr*nb_atoms)+interaction*Nr:nb_atoms*(Nrho+Nr*nb_atoms)+interaction*Nr+Nr]
                    interaction+=1
            rep[i,i,:] = data[nb_atoms*(Nrho+Nr*nb_atoms)+interaction*Nr:nb_atoms*(Nrho+Nr*nb_atoms)+interaction*Nr+Nr]
            interaction+=1
        eam.close()
        return source,parameters, F,f,rep
    else:
        print('Non supported eam file type')
        raise ValueError
            

def mix_eam(files,kind,method,f=[],rep_ab=[],alphas=[],betas=[]):
    """
    mix eam alloy files data set and compute the interspecies pair potential part using the 
    mean geometric value from each pure species 
    
    Parameters
    ----------
      files : array of strings
              Contain all the files to merge and mix
      kind : string
              kinf of eam. Supported eam/alloy, eam/fs
      method : string, (geometric,arithmetic,weighted,fitted)
              Method used to mix the pair interaction terms
              Available : Geometric average, arithmetic average, weighted arithmetic average
                  The Weighted arithmetic method is using the electron density function values of atom a and b to 
                  ponderate the pair potential between species a and b, 
                  rep_ab = 0.5*(fb/fa * rep_a + fa/fb * rep_b)
                  ref : X. W. Zhou, R. A. Johnson, and H. N. G. Wadley, Phys. Rev. B, 69, 144113 (2004)
                  Fitted method is to be used if the rep_ab has been previously fitted and is parse as rep_ab karg
      f : np.array of the fitted density term (for FS eam style)
      rep_ab : np.array of the fitted rep_ab term
      alphas : array of fitted alpha values for the fine tuned mixing. rep_ab = alpha_a*rep_a+alpha_b*rep_b
      betas : array of fitted values for the fine tuned mixing. f_ab = beta_00*rep_a+beta_01*rep_b
                                                                f_ba = beta_10*rep_a+beta_11*rep_b
    Returns
    -------
      sources : string
              Source informations or comment line for the file header
      parameters_mix : list of tuples
                [0] - array of str - atoms 
                [1] - array of int - atomic numbers
                [2] - array of float -atomic masses
                [3] - array of float - equilibrium lattice parameter
                [4] - array of str - crystal structure
                [5] - int - number of data point for embedded function
                [6] - int - number of data point for density and pair functions
                [7] - float - step size for the embedded function
                [8] - float - step size for the density and pair functions
                [9] - float - cutoff of the potentials
      F_ : array_like
          contain the tabulated values of the embedded functions
          shape = (nb atoms,nb atoms, nb of data points)
      f_ : array_like
          contain the tabulated values of the density functions
          shape = (nb atoms,nb atoms, nb of data points)
      rep_ : array_like
          contain the tabulated values of pair potential
          shape = (nb atoms,nb atoms, nb of data points)
    """

    nb_at = 0
    # Counting elements and repartition and select smallest tabulated set Nrho*drho // Nr*dr
    Nrho,drho,Nr,dr,cutoff = np.empty((len(files))),np.empty((len(files))),np.empty((len(files))),np.empty((len(files))),np.empty((len(files)))
    sources = ""
    if kind == "eam/alloy":
        for i,f_eam in enumerate(files):
            source,parameters, F,f,rep = read_eam(f_eam,kind="eam/alloy")
            sources+= source
            source += " "
            nb_at+=len(parameters[0])
            Nrho[i] = parameters[5]
            drho[i] = parameters[7]
            cutoff[i] = parameters[9]
            Nr[i] = parameters[6]
            dr[i] = parameters[8]
        # --- #
        max_cutoff = cutoff.argmax()
        max_prod = (Nrho*drho).argmax()
        max_prod_r = (Nr*dr).argmax()
        atnumber,atmass,crystallatt,crystal,atoms = np.empty(0),np.empty(0),np.empty(0),np.empty(0).astype(np.str),np.empty(0).astype(np.str)
        Nr_ = Nr[max_prod_r]
        dr_ = ((Nr*dr).max())/Nr_
        Nrho_ = Nrho[max_prod]
        drho_ = ((Nrho*drho).max())/Nrho_
        
        if Nr_ > 2000:
          Nr_ = 2000   # reduce
          dr_ = ((Nr*dr).max())/Nr_
        if Nrho_ > 2000:
          Nrho_ = 2000 # reduce
          drho_ = ((Nrho*drho).max())/Nrho_
        F_,f_,rep_ = np.empty((nb_at,Nrho_)),np.empty((nb_at,Nr_)),np.empty((nb_at,nb_at,Nr_))
        at = 0
        for i,f_eam in enumerate(files):
            source,parameters, F,f,rep = read_eam(f_eam,kind="eam/alloy")
            atoms = np.append(atoms,parameters[0])
            atnumber = np.append(atnumber,parameters[1])
            atmass = np.append(atmass,parameters[2])
            crystallatt = np.append(crystallatt,parameters[3])
            crystal = np.append(crystal,parameters[4])
            for j in range(len(parameters[0])):
                F_[at,:] = interpolate.InterpolatedUnivariateSpline(np.linspace(0,Nrho[i]*drho[i],Nrho[i]),F[j,:])(np.linspace(0,Nrho_*drho_,Nrho_))
                f_[at,:] = interpolate.InterpolatedUnivariateSpline(np.linspace(0,Nr[i]*dr[i],Nr[i]),f[j,:])(np.linspace(0,Nr_*dr_,Nr_))
                rep_[at,at,:] = interpolate.InterpolatedUnivariateSpline(np.linspace(0,Nr[i]*dr[i],Nr[i]),rep[j,j,:])(np.linspace(0,Nr_*dr_,Nr_))
                at+=1
        # mixing repulsive part
        old_err_state = np.seterr(divide='raise')
        ignored_states = np.seterr(**old_err_state)
        for i in range(nb_at):
            for j in range(nb_at):
                if j < i :
                    if method == "geometric":
                        rep_[i,j,:] = (rep_[i,i,:]*rep_[j,j,:])**0.5
                    if method == "arithmetic":
                        if alphas:
                            rep_[i,j,:] = alphas[i]*rep_[i,i,:]+alphas[j]*rep_[j,j,:]
                        else:
                            rep_[i,j,:] = 0.5*(rep_[i,i,:]+rep_[j,j,:])
                    if method == "weighted":
                        rep_[i,j,:] = 0.5*(np.divide(f_[j,:],f_[i,:])*rep_[i,i,:]+np.divide(f_[i,:],f_[j,:])*rep_[j,j,:])
                    if method == "fitted":
                        rep_ab[np.isnan(rep_ab)] = 0
                        rep_ab[np.isinf(rep_ab)] = 0
                        rep_[i,j,:] = interpolate.InterpolatedUnivariateSpline(np.linspace(0,max(Nr*dr),rep_ab.shape[0]),rep_ab)(np.linspace(0,Nr_*dr_,Nr_))
                    rep_[i,j,:][np.isnan(rep_[i,j,:])] = 0
                    rep_[i,j,:][np.isinf(rep_[i,j,:])] = 0
    elif kind == "eam/fs":
        for i,f_eam in enumerate(files):
            source,parameters, F,f,rep = read_eam(f_eam,kind="eam/alloy")
            sources+= source
            source += " "
            nb_at+=len(parameters[0])
            Nrho[i] = parameters[5]
            drho[i] = parameters[7]
            cutoff[i] = parameters[9]
            Nr[i] = parameters[6]
            dr[i] = parameters[8]
        # --- #
        max_cutoff = cutoff.argmax()
        max_prod = (Nrho*drho).argmax()
        max_prod_r = (Nr*dr).argmax()
        atnumber,atmass,crystallatt,crystal,atoms = np.empty(0),np.empty(0),np.empty(0),np.empty(0).astype(np.str),np.empty(0).astype(np.str)
        Nr_ = Nr[max_prod_r]
        dr_ = ((Nr*dr).max())/Nr_
        Nrho_ = Nrho[max_prod]
        drho_ = ((Nrho*drho).max())/Nrho_
        
        if Nr_ > 2000:
          Nr_ = 2000   # reduce
          dr_ = ((Nr*dr).max())/Nr_
        if Nrho_ > 2000:
          Nrho_ = 2000 # reduce
          drho_ = ((Nrho*drho).max())/Nrho_
        F_,f_,rep_ = np.empty((nb_at,Nrho_)),np.empty((nb_at,nb_at,Nr_)),np.empty((nb_at,nb_at,Nr_))
        at = 0
        for i,f_eam in enumerate(files):
            source,parameters, F,f,rep = read_eam(f_eam,kind="eam/alloy")
            atoms = np.append(atoms,parameters[0])
            atnumber = np.append(atnumber,parameters[1])
            atmass = np.append(atmass,parameters[2])
            crystallatt = np.append(crystallatt,parameters[3])
            crystal = np.append(crystal,parameters[4])
            for j in range(len(parameters[0])):
                F_[at,:] = interpolate.InterpolatedUnivariateSpline(np.linspace(0,Nrho[i]*drho[i],Nrho[i]),F[j,:])(np.linspace(0,Nrho_*drho_,Nrho_))
                f_[at,at,:] = interpolate.InterpolatedUnivariateSpline(np.linspace(0,Nr[i]*dr[i],Nr[i]),f[j,:])(np.linspace(0,Nr_*dr_,Nr_))
                rep_[at,at,:] = interpolate.InterpolatedUnivariateSpline(np.linspace(0,Nr[i]*dr[i],Nr[i]),rep[j,j,:])(np.linspace(0,Nr_*dr_,Nr_))
                at+=1
        # mixing density part
        old_err_state = np.seterr(divide='raise')
        ignored_states = np.seterr(**old_err_state)
        for i in range(nb_at):
            for j in range(nb_at):
                if i!=j:
                    if method == "geometric":
                        f_[i,j,:] = (f_[i,i,:]*f_[j,j,:])**0.5
                    if method == "arithmetic":
                        if betas.any():
                            f_[i,j,:] = betas[i,i]*f_[i,i,:]+betas[i,j]*f_[j,j,:]
                        else:
                            f_[i,j,:] = 0.5*(f_[i,i,:]+f_[j,j,:])
                    if method == "fitted":
                        f_ab[np.isnan(f_ab)] = 0
                        f_ab[np.isinf(f_ab)] = 0
                        f_[i,j,:] = interpolate.InterpolatedUnivariateSpline(np.linspace(0,max(Nr*dr),rep_ab.shape[0]),rep_ab)(np.linspace(0,Nr_*dr_,Nr_))
                    f_[i,j,:][np.isnan(f_[i,j,:])] = 0
                    f_[i,j,:][np.isinf(f_[i,j,:])] = 0
        # mixing repulsive part
        for i in range(nb_at):
            for j in range(nb_at):
                if j < i :
                    if method == "geometric":
                        rep_[i,j,:] = (rep_[i,i,:]*rep_[j,j,:])**0.5
                    if method == "arithmetic":
                        if alphas:
                            rep_[i,j,:] = alphas[i]*rep_[i,i,:]+alphas[j]*rep_[j,j,:]
                        else:
                            rep_[i,j,:] = 0.5*(rep_[i,i,:]+rep_[j,j,:])
                    if method == "fitted":
                        rep_ab[np.isnan(rep_ab)] = 0
                        rep_ab[np.isinf(rep_ab)] = 0
                        rep_[i,j,:] = interpolate.InterpolatedUnivariateSpline(np.linspace(0,max(Nr*dr),rep_ab.shape[0]),rep_ab)(np.linspace(0,Nr_*dr_,Nr_))
                    rep_[i,j,:][np.isnan(rep_[i,j,:])] = 0
                    rep_[i,j,:][np.isinf(rep_[i,j,:])] = 0
    else:
        print('Non supported eam file type')
        raise ValueError
                
    parameters_mix = EAMParameters(atoms, atnumber, atmass,crystallatt,crystal, Nrho_,Nr_, drho_, dr_, cutoff[max_cutoff])
    return sources, parameters_mix, F_, f_, rep_
      
def write_eam(source, parameters, F, f, rep,out_file,kind="eam"):
    """
    Write an eam lammps format file 
    http://lammps.sandia.gov/doc/pair_eam.html
    
    Parameters
    ----------
    source : string
          Source information or comment line for the file header
    parameters : list of tuples
                [0] - array of str - atoms (ONLY FOR eam/alloy and eam/fs, EMPTY for eam)
                [1] - array of int - atomic numbers
                [2] - array of float -atomic masses
                [3] - array of float - equilibrium lattice parameter
                [4] - array of str - crystal structure
                [5] - int - number of data point for embedded function
                [6] - int - number of data point for density and pair functions
                [7] - float - step size for the embedded function
                [8] - float - step size for the density and pair functions
                [9] - float - cutoff of the potentials
    F : array_like
        contain the tabulated values of the embedded functions
        shape = (nb of data points)
    f : array_like
        contain the tabulated values of the density functions
        shape = (nb of data points)
    rep : array_like
        contain the tabulated values of pair potential
        shape = (nb of data points)
    out_file : string
              output file name for the eam alloy potential file
    Returns
    -------
      
    """
  
    atoms, atnumber, atmass, crystallatt, crystal = parameters[0:5]
    Nrho, Nr, drho, dr, cutoff = parameters[5:10]
    
    if kind == "eam":
        # parameters unpacked
        atline = "%i %f %f %s"%(int(atnumber),float(atmass),float(crystallatt),str(crystal))
        parameterline = '%i\t%.16e\t%i\t%.16e\t%.10e'%(int(Nrho),float(drho),int(Nr),float(dr),float(cutoff))
        potheader = "# EAM potential from : # %s \n %s \n %s"%(source,atline,parameterline)
        # --- Writing new EAM alloy pot file --- #
        # write header and file parameters
        potfile = open(out_file,'wb')
        # write F and f tables
        np.savetxt(potfile, F, fmt='%.16e',header=potheader,comments='')
        np.savetxt(potfile, f, fmt='%.16e')
        # write pair interactions tables
        np.savetxt(potfile, rep, fmt='%.16e')
        potfile.close()  
    elif kind == "eam/alloy":
        nb_atoms = len(atoms)
        # parameters unpacked
        potheader = "# Mixed EAM alloy potential from :\n# %s \n# \n"%(source)
        # --- Writing new EAM alloy pot file --- #
        potfile = open(out_file,'wb')
        # write header and file parameters
        np.savetxt(potfile,atoms,fmt="%s",newline=' ', header=potheader+str(nb_atoms),footer='\n%i\t%e\t%i\t%e\t%e\n'%(Nrho,drho,Nr,dr,cutoff), comments='')
        # write F and f tables
        [np.savetxt(potfile,np.append(F[i,:],f[i,:]),fmt="%.16e",header='%i\t%f\t%f\t%s'%(atnumber[i],atmass[i],crystallatt[i],crystal[i]),comments='') for i in range(nb_atoms)]
        # write pair interactions tables
        [[np.savetxt(potfile,rep[i,j,:],fmt="%.16e") for j in range(rep.shape[0]) if j <= i] for i in range(rep.shape[0])]
        potfile.close() 
    elif kind == "eam/fs":
        nb_atoms = len(atoms)
        # parameters unpacked
        potheader = "# Mixed EAM fs potential from :\n# %s \n# \n"%(source)
        # --- Writing new EAM alloy pot file --- #
        potfile = open(out_file,'wb')
        # write header and file parameters
        np.savetxt(potfile,atoms,fmt="%s",newline=' ', header=potheader+str(nb_atoms),footer='\n%i\t%e\t%i\t%e\t%e\n'%(Nrho,drho,Nr,dr,cutoff), comments='')
        # write F and f tables
        [np.savetxt(potfile,np.append(F[i,:],f[i,:,:].flatten()),fmt="%.16e",header='%i\t%f\t%f\t%s'%(atnumber[i],atmass[i],crystallatt[i],crystal[i]),comments='') for i in range(nb_atoms)]
        # write pair interactions tables
        [[np.savetxt(potfile,rep[i,j,:],fmt="%.16e") for j in range(rep.shape[0]) if j <= i] for i in range(rep.shape[0])]
        potfile.close() 
    else:
        print('Non supported eam file type')
        raise ValueError



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

'''
Compute the spatial correlation of a given function. Distances larger
than a cut-off are calculated by fourier transformation, while shorter
distances are calculated directly.
coords.shape=(n_atoms,3)
cell_vectors=[[x1,y1,z1],[x2,y2,z2],[x3,y3,z3]]

options:
#1)dimensions of correlation [dim=]
    a: along all 3 dimensions -> total distance: None (default)
    b: along only one dimension, ignoring other components: 0..2
#2)mode to assign the atomic values to the FFT grid points [delta=]
    a: assign value to the nearest grid point: simple (default)
    b: assign value to the 8 nearest grid points, distributed
        proportional to their distance: else
#3)nomalisation by variance of values [norm=]
    a: off: False (default)
    b: on: True
'''

from __future__ import division

import numpy as np
from math import floor, ceil
from matscipy.neighbours import neighbour_list
from ase import Atoms

def betrag(vec):
    return np.sqrt((vec**2).sum())
def max_rad(cell_vectors):
    #calculate length cutoff from cell vectors
    x = cell_vectors[:,0]
    y = cell_vectors[:,1]
    z = cell_vectors[:,2]
    r = np.zeros(3)
    nor = np.zeros(3)

    nor = np.cross(y,z)
    r[0] = np.abs((x*nor).sum()/betrag(x)/betrag(nor))*betrag(x)/2.0
    nor = np.cross(z,x)
    r[1] = np.abs((y*nor).sum()/betrag(y)/betrag(nor))*betrag(y)/2.0
    nor = np.cross(x,y)
    r[2] = np.abs((z*nor).sum()/betrag(z)/betrag(nor))*betrag(z)/2.0
    
    return r.min()


def spatial_correlation_function(atoms, values, length_cutoff=None,
                                 output_gridsize=None, FFT_cutoff=None,
                                 approx_FFT_gridsize=None, dim=None,
                                 delta='simple', norm=False):
    # Make sure values are floats
    values = np.asarray(values, dtype=float)

    xyz = atoms.get_positions()
    abc = atoms.get_scaled_positions() % 1.0
    cell_vectors = atoms.cell.T
    n_atoms = len(xyz)

    if length_cutoff is None:
        length_cutoff = np.floor(max_rad(cell_vectors))

    if FFT_cutoff is None:
        FFT_cutoff = 7.5

    if output_gridsize is None:
        output_gridsize = 0.1

    if approx_FFT_gridsize is None:
        approx_FFT_gridsize = 1.0

    n_lattice_points = np.array(np.ceil(cell_vectors.diagonal()
                                        /approx_FFT_gridsize),
                                dtype=int)
    FFT_gridsize = cell_vectors.diagonal()/n_lattice_points

    if delta == 'simple':
        # calc lattice values (add to nearest lattice point)
        Q = np.zeros(shape=(n_lattice_points))
        for _abc, _q in zip(abc, values):
            x,y,z = np.array(_abc*n_lattice_points, dtype=int) \
                             %n_lattice_points
            Q[x,y,z] += _q
    else:
        # proportional distribution on 8 neightbor points
        Q = np.zeros(shape=(n_lattice_points))
        a1, a2, a3 = cell_vectors.T
        for _abc, _q in zip(abc, q):
            x,y,z = _abc*(n_lattice_points-1)
            aes = np.array([np.floor(x),np.ceil(x)] \
                          ).reshape(-1, 1, 1, 1)/(n_lattice_points[0]-1)
            bes = np.array([np.floor(y),np.ceil(y)] \
                          ).reshape( 1,-1, 1, 1)/(n_lattice_points[1]-1)
            ces = np.array([np.floor(z),np.ceil(z)] \
                          ).reshape( 1, 1,-1, 1)/(n_lattice_points[2]-1)
            octo = (aes*a1.reshape(1,1,1,-1) \
                    +bes*a2.reshape(1,1,1,-1) \
                    +ces*a3.reshape(1,1,1,-1)) \
                    -cartesianN(_abc,cell_vectors).reshape(1,1,1,-1)
            octo = 1./(np.sqrt((octo**2).sum(axis=3)))
            Q[np.floor(x):np.ceil(x)+1,np.floor(y):np.ceil(y)+1,
              np.floor(z):np.ceil(z)+1] += octo/octo.sum()*_q

    # FFT
    Q_schlange = np.fft.fftn(Q)
    C_schlange = Q_schlange*Q_schlange.conjugate()
    C = np.fft.ifftn(C_schlange)*n_lattice_points.prod() \
        /n_atoms/n_atoms
    C = np.fft.ifftshift(C)

    if dim is None:
        # distance mapping (for floor/ceil convention see *i*fftshift
        # definition)
        a = np.reshape(np.arange(-floor(n_lattice_points[0]/2.),
                       ceil(n_lattice_points[0]/2.),1)
                       /n_lattice_points[0],(-1, 1, 1, 1))
        b = np.reshape(np.arange(-floor(n_lattice_points[1]/2.),
                       ceil(n_lattice_points[1]/2.),1)
                       /n_lattice_points[1],( 1,-1, 1, 1))
        c = np.reshape(np.arange(-floor(n_lattice_points[2]/2.),
                       ceil(n_lattice_points[2]/2.),1)
                       /n_lattice_points[2],( 1, 1,-1, 1))
        a1, a2, a3 = cell_vectors.T

        r = a*a1.reshape(1,1,1,-1)+b*a2.reshape(1,1,1,-1) \
            +c*a3.reshape(1,1,1,-1)
        dist = np.sqrt((r**2).sum(axis=3))
    elif 0 <= dim <3:
        # directional SCFs
        # for floor/ceil convention see *i*fftshift definition
        a = np.reshape(np.arange(-floor(n_lattice_points[0]/2.),
                                 ceil(n_lattice_points[0]/2.),1)
                       /n_lattice_points[0],(-1, 1, 1, 1))
        b = np.reshape(np.arange(-floor(n_lattice_points[1]/2.),
                                 ceil(n_lattice_points[1]/2.),1)
                       /n_lattice_points[1],( 1,-1, 1, 1))
        c = np.reshape(np.arange(-floor(n_lattice_points[2]/2.),
                                 ceil(n_lattice_points[2]/2.),1)
                       /n_lattice_points[2],( 1, 1,-1, 1))
        a1, a2, a3 = cell_vectors.T
        r = a*a1.reshape(1,1,1,-1) +b*a2.reshape(1,1,1,-1) \
            +c*a3.reshape(1,1,1,-1)
        dist = np.abs(r[:,:,:,dim])  # use indices to access directions
    else:
        print('invalid correlation direction: '+str(dim))
        sys.exit()

    nbins = int(length_cutoff/output_gridsize)
    bins = np.arange(0, length_cutoff+length_cutoff/nbins,
                     length_cutoff/nbins)
    SCF, edges = np.histogram(np.ravel(dist), bins=bins,
                              weights=np.ravel(np.real(C)))
    n, edges = np.histogram(np.reshape(dist,(-1,1)), bins=bins)
    n[n==0] = 1
    SCF /= n
    # Alternative to the above three lines:
    # SCF *= atoms.get_volume()/np.prod(n_lattice_points) / slice_volume
    if norm:
        v_2_mean = (values**2).mean()
        v_mean_2 = (values.mean())**2
        SCF = (SCF-v_mean_2)/(v_2_mean-v_mean_2)

    return SCF, (edges[1:]+edges[:-1])/2



def spatial_correlation_function_near(atoms, values, gridsize=None,
                                      cutoff=None, norm=False):
    if gridsize is None:
        gridsize = 0.1

    if cutoff is None:
        cutoff = 7.5

    # close range exact calculation
    nbins = int(cutoff/gridsize)+1
    index1,index2,dist = neighbour_list('ijd', atoms, cutoff=cutoff)
    SCF_near, edges = np.histogram(dist, bins=bins,
                                   weights=values[index1]
                                   *values[index2])
    slice_volume = 4*np.pi/3 * (edges[1:]**3-edges[:-1]**3)
    SCF_near *= atoms.get_volume()/n_atoms**2 / slice_volume
    if norm:
        v_2_mean = (values**2).mean()
        v_mean_2 = (values.mean())**2
        SCF_near = (SCF_near-v_mean_2)/(v_2_mean-v_mean_2)

    return SCF_near, (edges[1:]+edges[:-1])/2
